"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
exports.syncCycleData = exports.compareWithOldCyclesData = exports.compareWithOldReceiptsData = exports.compareWithOldOriginalTxsData = exports.syncCyclesAndTxsDataBetweenCycles = exports.syncCyclesAndTxsData = exports.syncOriginalTxsByCycle = exports.syncOriginalTxs = exports.syncReceiptsByCycle = exports.syncReceipts = exports.syncCyclesBetweenCycles = exports.syncCyclesAndNodeListV2 = exports.syncCyclesAndNodeList = exports.buildNodeListFromStoredCycle = exports.syncGenesisAccountsFromConsensor = exports.syncGenesisTransactionsFromArchiver = exports.syncGenesisAccountsFromArchiver = exports.getTotalDataFromArchivers = exports.checkActiveStatus = exports.checkJoinStatus = exports.getCycleDuration = exports.sendActiveRequest = exports.sendLeaveRequest = exports.submitJoin = exports.joinNetwork = exports.calcIncomingTimes = exports.clearDataSenders = exports.sendDataRequest = exports.subscribeNodeFromThisSubset = exports.subscribeConsensorsByConsensusRadius = exports.createNodesGroupByConsensusRadius = exports.createDataTransferConnection = exports.getConsensusRadius = exports.addDataSender = exports.createContactTimeout = exports.subscribeNodeForDataTransfer = exports.replaceDataSender = exports.emitter = exports.dataSenders = exports.clearCombinedAccountsData = exports.collectCycleData = exports.initSocketClient = exports.unsubscribeDataSender = exports.createDataRequest = exports.DataRequestTypes = exports.nodesPerEdge = exports.nodesPerConsensusGroup = exports.currentConsensusRadius = exports.combineAccountsData = exports.socketClients = void 0;
exports.scoreCert = void 0;
const events_1 = require("events");
const Crypto = require("../Crypto");
const NodeList = require("../NodeList");
const Cycles = require("./Cycles");
const Cycles_1 = require("./Cycles");
const CycleParser_1 = require("./CycleParser");
const State = require("../State");
const P2P = require("../P2P");
const Utils = require("../Utils");
const Config_1 = require("../Config");
const lib_types_1 = require("@shardeum-foundation/lib-types");
const Logger = require("../Logger");
const nestedCounters_1 = require("../profiler/nestedCounters");
const Collector_1 = require("./Collector");
const CycleDB = require("../dbstore/cycles");
const ReceiptDB = require("../dbstore/receipts");
const OriginalTxDB = require("../dbstore/originalTxsData");
const StateMetaData = require("../archivedCycle/StateMetaData");
const sync_v2_1 = require("../sync-v2");
const API_1 = require("../API");
const ioclient = require("socket.io-client");
const P2P_1 = require("../P2P");
const Utils_1 = require("../Utils");
const lib_types_2 = require("@shardeum-foundation/lib-types");
const cycleRecordsCache_1 = require("../cache/cycleRecordsCache");
const general_1 = require("../utils/general");
const customHttpFunctions_1 = require("../utils/customHttpFunctions");
const archiverLogging_1 = require("../profiler/archiverLogging");
const lib_types_3 = require("@shardeum-foundation/lib-types");
class ValidationTracker {
    constructor() {
        this.seen = new Set();
        this.breadcrumbs = [];
        this.MAX_ENTRIES = 1000;
    }
    add(breadcrumb) {
        const key = `${breadcrumb.cycle.marker}:${breadcrumb.cycle.previous}`;
        if (this.seen.has(key)) {
            return;
        }
        if (this.breadcrumbs.length >= this.MAX_ENTRIES) {
            const oldest = this.breadcrumbs.shift();
            this.seen.delete(`${oldest.cycle.marker}:${oldest.cycle.previous}`);
        }
        Logger.mainLogger.warn('[ValidationTracker] Certificate validation failed', {
            cycle: breadcrumb.cycle,
        });
        this.seen.add(key);
        this.breadcrumbs.push(breadcrumb);
    }
}
const validationTracker = new ValidationTracker();
exports.socketClients = new Map();
exports.combineAccountsData = {
    accounts: [],
    receipts: [],
};
const forwardGenesisAccounts = true;
exports.currentConsensusRadius = 0;
exports.nodesPerConsensusGroup = 0;
exports.nodesPerEdge = 0;
let subsetNodesMapByConsensusRadius = new Map();
const maxCyclesInCycleTracker = 5;
const receivedCycleTracker = {};
const QUERY_TIMEOUT_MAX = 30; // 30seconds
const { MAX_ACCOUNTS_PER_REQUEST, MAX_RECEIPTS_PER_REQUEST, MAX_ORIGINAL_TXS_PER_REQUEST, MAX_CYCLES_PER_REQUEST, MAX_BETWEEN_CYCLES_PER_REQUEST, } = Config_1.config.REQUEST_LIMIT;
const GENESIS_ACCOUNTS_CYCLE_RANGE = {
    startCycle: 0,
    endCycle: 5,
};
var DataRequestTypes;
(function (DataRequestTypes) {
    DataRequestTypes["SUBSCRIBE"] = "SUBSCRIBE";
    DataRequestTypes["UNSUBSCRIBE"] = "UNSUBSCRIBE";
})(DataRequestTypes = exports.DataRequestTypes || (exports.DataRequestTypes = {}));
function createDataRequest(type, lastData, recipientPk) {
    return Crypto.tag({
        type,
        lastData,
    }, recipientPk);
}
exports.createDataRequest = createDataRequest;
async function unsubscribeDataSender(publicKey) {
    Logger.mainLogger.debug('Disconnecting previous connection', publicKey);
    const sender = exports.dataSenders.get(publicKey);
    if (sender) {
        // Clear contactTimeout associated with this sender
        if (sender.contactTimeout) {
            clearTimeout(sender.contactTimeout);
            sender.contactTimeout = null;
        }
        sendDataRequest(sender.nodeInfo, DataRequestTypes.UNSUBSCRIBE);
        // Delete sender from dataSenders
        exports.dataSenders.delete(publicKey);
    }
    const socketClient = exports.socketClients.get(publicKey);
    if (socketClient) {
        socketClient.emit('UNSUBSCRIBE', Config_1.config.ARCHIVER_PUBLIC_KEY);
        socketClient.close();
        exports.socketClients.delete(publicKey);
    }
    nestedCounters_1.nestedCountersInstance.countEvent('archiver', 'remove_data_sender');
    Logger.mainLogger.debug('Subscribed dataSenders', exports.dataSenders.size, 'Connected socketClients', exports.socketClients.size);
    if (Config_1.config.VERBOSE)
        Logger.mainLogger.debug('Subscribed dataSenders', exports.dataSenders.keys(), 'Connected socketClients', exports.socketClients.keys());
}
exports.unsubscribeDataSender = unsubscribeDataSender;
function initSocketClient(node) {
    if (Config_1.config.VERBOSE)
        Logger.mainLogger.debug('Node Info to socket connect', node);
    try {
        const socketClient = ioclient.connect(`http://${node.ip}:${node.port}`, {
            query: {
                data: JSON.stringify(Crypto.sign({
                    publicKey: State.getNodeInfo().publicKey,
                    timestamp: Date.now(),
                    intendedConsensor: node.publicKey,
                })),
            },
        });
        exports.socketClients.set(node.publicKey, socketClient);
        socketClient.on('connect', () => {
            Logger.mainLogger.debug(`✅ New Socket Connection to consensus node ${node.ip}:${node.port} is made`);
            if (Config_1.config.VERBOSE)
                Logger.mainLogger.debug('Connected node', node);
            if (Config_1.config.VERBOSE)
                Logger.mainLogger.debug('Init socketClients', exports.socketClients.size, exports.dataSenders.size);
        });
        socketClient.once('disconnect', async () => {
            Logger.mainLogger.debug(`Connection request is refused by the consensor node ${node.ip}:${node.port}`);
        });
        socketClient.on('DATA', (data) => {
            const newData = lib_types_2.Utils.safeJsonParse(data);
            if (!newData || !newData.responses)
                return;
            if (newData.recipient !== State.getNodeInfo().publicKey) {
                Logger.mainLogger.debug('This data is not meant for this archiver');
                return;
            }
            // If tag is invalid, dont keepAlive, END
            if (Crypto.authenticate(newData) === false) {
                Logger.mainLogger.debug('This data cannot be authenticated');
                unsubscribeDataSender(node.publicKey);
                return;
            }
            if (Config_1.config.experimentalSnapshot) {
                // Get sender entry
                let sender = exports.dataSenders.get(newData.publicKey);
                // If no sender entry, remove publicKey from senders, END
                if (!sender) {
                    Logger.mainLogger.error('This sender is not in the subscribed nodes list', newData.publicKey);
                    // unsubscribeDataSender(newData.publicKey)
                    return;
                }
                // Clear senders contactTimeout, if it has one
                if (sender.contactTimeout) {
                    if (Config_1.config.VERBOSE)
                        Logger.mainLogger.debug('Clearing contact timeout.');
                    clearTimeout(sender.contactTimeout);
                    sender.contactTimeout = null;
                    nestedCounters_1.nestedCountersInstance.countEvent('archiver', 'clear_contact_timeout');
                }
                if (Config_1.config.VERBOSE)
                    console.log('DATA', sender.nodeInfo.publicKey, sender.nodeInfo.ip, sender.nodeInfo.port);
                if (newData.responses && newData.responses.ORIGINAL_TX_DATA) {
                    if (Config_1.config.VERBOSE)
                        Logger.mainLogger.debug('ORIGINAL_TX_DATA', sender.nodeInfo.publicKey, sender.nodeInfo.ip, sender.nodeInfo.port, newData.responses.ORIGINAL_TX_DATA.length);
                    // gracefully ignoring since it is now coupled with the receipt flow
                    // storeOriginalTxData(
                    //   newData.responses.ORIGINAL_TX_DATA,
                    //   sender.nodeInfo.ip + ':' + sender.nodeInfo.port,
                    //   config.saveOnlyGossipData
                    // )
                }
                if (newData.responses && newData.responses.RECEIPT) {
                    if (Config_1.config.VERBOSE)
                        Logger.mainLogger.debug('RECEIPT', sender.nodeInfo.publicKey, sender.nodeInfo.ip, sender.nodeInfo.port, newData.responses.RECEIPT.length);
                    (0, Collector_1.storeReceiptData)(newData.responses.RECEIPT, sender.nodeInfo.ip + ':' + sender.nodeInfo.port, true, Config_1.config.saveOnlyGossipData);
                }
                if (newData.responses && newData.responses.CYCLE) {
                    collectCycleData(newData.responses.CYCLE, sender.nodeInfo.ip + ':' + sender.nodeInfo.port, 'data-sender');
                }
                if (newData.responses && newData.responses.ACCOUNT) {
                    if ((0, Cycles_1.getCurrentCycleCounter)() > GENESIS_ACCOUNTS_CYCLE_RANGE.endCycle) {
                        Logger.mainLogger.error('Account data is not meant to be received after the genesis accounts cycle range', (0, Cycles_1.getCurrentCycleCounter)());
                        unsubscribeDataSender(sender.nodeInfo.publicKey);
                        return;
                    }
                    if (Cycles.currentNetworkMode !== 'forming' ||
                        NodeList.byPublicKey.size > 1 ||
                        !NodeList.byPublicKey.has(sender.nodeInfo.publicKey)) {
                        Logger.mainLogger.error('Account data is not meant to be received by the first validator', `Number of nodes in the network ${NodeList.byPublicKey.size}`);
                        unsubscribeDataSender(sender.nodeInfo.publicKey);
                        return;
                    }
                    Logger.mainLogger.debug(`RECEIVED ACCOUNTS DATA FROM ${sender.nodeInfo.ip}:${sender.nodeInfo.port}`);
                    nestedCounters_1.nestedCountersInstance.countEvent('genesis', 'accounts', 1);
                    if (!forwardGenesisAccounts) {
                        Logger.mainLogger.debug('Genesis Accounts To Sycn', newData.responses.ACCOUNT);
                        syncGenesisAccountsFromConsensor(newData.responses.ACCOUNT, sender.nodeInfo);
                    }
                    else {
                        if (Collector_1.storingAccountData) {
                            Logger.mainLogger.debug('Storing Account Data');
                            let newCombineAccountsData = { ...exports.combineAccountsData };
                            if (newData.responses.ACCOUNT.accounts)
                                newCombineAccountsData.accounts = [
                                    ...newCombineAccountsData.accounts,
                                    ...newData.responses.ACCOUNT.accounts,
                                ];
                            if (newData.responses.ACCOUNT.receipts)
                                newCombineAccountsData.receipts = [
                                    ...newCombineAccountsData.receipts,
                                    ...newData.responses.ACCOUNT.receipts,
                                ];
                            exports.combineAccountsData = { ...newCombineAccountsData };
                            newCombineAccountsData = {
                                accounts: [],
                                receipts: [],
                            };
                        }
                        else
                            (0, Collector_1.storeAccountData)(newData.responses.ACCOUNT);
                    }
                }
                // Set new contactTimeout for sender. Postpone sender removal because data is still received from consensor
                nestedCounters_1.nestedCountersInstance.countEvent('archiver', 'postpone_contact_timeout');
                // To make sure that the sender is still in the subscribed list
                sender = exports.dataSenders.get(newData.publicKey);
                if (sender)
                    sender.contactTimeout = createContactTimeout(sender.nodeInfo.publicKey, 'This timeout is created after processing data');
            }
        });
    }
    catch (error) {
        console.error('Error occurred during socket connection:', error);
    }
}
exports.initSocketClient = initSocketClient;
function collectCycleData(cycleData, senderInfo, source) {
    const startTime = Date.now();
    const operationId = archiverLogging_1.ArchiverLogging.generateOperationId();
    Logger.mainLogger.debug(`collectCycleData: Processing ${cycleData.length} cycles from ${senderInfo}, source: ${source}`);
    nestedCounters_1.nestedCountersInstance.countEvent('collectCycleData', 'cycles_received', cycleData.length);
    nestedCounters_1.nestedCountersInstance.countEvent('collectCycleData', 'source_' + source, 1);
    archiverLogging_1.ArchiverLogging.logDataSync({
        sourceArchiver: senderInfo,
        targetArchiver: Config_1.config.ARCHIVER_IP,
        cycle: 0,
        dataType: 'CYCLE_RECORD',
        dataHash: '',
        status: 'STARTED',
        operationId,
        metrics: {
            duration: 0,
            dataSize: JSON.stringify(cycleData).length,
        },
    });
    // check if the sender is in the nodelists
    if (NodeList.activeListByIdSorted.length > 0) {
        const [ip, port] = senderInfo.split(':');
        const isInActiveNodes = NodeList.activeListByIdSorted.some((node) => node.ip === ip && node.port.toString() === port);
        const isInActiveArchivers = State.activeArchivers.some((archiver) => archiver.ip === ip && archiver.port.toString() === port);
        if (!isInActiveNodes && !isInActiveArchivers) {
            nestedCounters_1.nestedCountersInstance.countEvent('collectCycleData', 'sender_not_active', 1);
            Logger.mainLogger.warn(`collectCycleData: Ignoring cycle data from non-active node: ${senderInfo}`);
            archiverLogging_1.ArchiverLogging.logDataSync({
                sourceArchiver: senderInfo,
                targetArchiver: Config_1.config.ARCHIVER_IP,
                cycle: 0,
                dataType: 'CYCLE_RECORD',
                dataHash: '',
                status: 'ERROR',
                operationId,
                metrics: {
                    duration: Date.now() - startTime,
                    dataSize: JSON.stringify(cycleData).length,
                },
                error: 'Sender not in active nodes or archivers',
            });
            return;
        }
    }
    for (const cycle of cycleData) {
        Logger.mainLogger.debug(`collectCycleData: Processing cycle ${cycle.counter}, marker: ${cycle.marker}`);
        if (receivedCycleTracker[cycle.counter]?.saved === true) {
            nestedCounters_1.nestedCountersInstance.countEvent('collectCycleData', 'cycle_already_saved_' + cycle.mode, 1);
            Logger.mainLogger.debug(`collectCycleData: Cycle ${cycle.counter} already saved, skipping`);
            archiverLogging_1.ArchiverLogging.logDataSync({
                sourceArchiver: senderInfo,
                targetArchiver: Config_1.config.ARCHIVER_IP,
                cycle: cycle.counter,
                dataType: 'CYCLE_RECORD',
                dataHash: cycle.marker,
                status: 'COMPLETE',
                operationId,
                metrics: {
                    duration: Date.now() - startTime,
                    dataSize: JSON.stringify(cycle).length,
                },
            });
            break;
        }
        nestedCounters_1.nestedCountersInstance.countEvent('collectCycleData', 'process_cycle_' + cycle.mode, 1);
        // since we can trust archivers and archiver only gossip after they have verified the cycleData
        // we can just call processCycles here
        if (source === 'archiver') {
            nestedCounters_1.nestedCountersInstance.countEvent('collectCycleData', 'direct_process_from_archiver', 1);
            Logger.mainLogger.debug(`collectCycleData: Processing cycle ${cycle.counter} from archiver directly`);
            (0, Cycles_1.processCycles)([cycle]);
            continue;
        }
        let receivedCertSigners = [];
        if (NodeList.activeListByIdSorted.length > 0) {
            const certSigners = receivedCycleTracker[cycle.counter]?.[cycle.marker]?.['certSigners'] ?? new Set();
            try {
                // need to get the hash(marker) of the cycle as it was in q3/q4 when the certs were made and compared
                Logger.mainLogger.debug(`collectCycleData: Original cycle data: ${lib_types_3.Utils.safeStringify(cycle)}`);
                const cycleCopy = getRecordWithoutPostQ3Changes(cycle);
                const computedMarker = Cycles.computeCycleMarker(cycleCopy);
                Logger.mainLogger.debug(`collectCycleData: cycle copy ${lib_types_3.Utils.safeStringify(cycleCopy)}`);
                Logger.mainLogger.debug(`collectCycleData: Computed marker for cycle ${cycle.counter}: ${computedMarker}, original marker: ${cycle.marker}`);
                Logger.mainLogger.debug(`collectCycleData: Validating ${cycle.certificates?.length || 0} certificates for cycle ${cycle.counter}`);
                const validateCertsResult = validateCerts(cycle.certificates, certSigners, computedMarker, cycleCopy);
                if (validateCertsResult === false) {
                    nestedCounters_1.nestedCountersInstance.countEvent('collectCycleData', 'certificate_validation_failed_' + cycle.mode, 1);
                    Logger.mainLogger.warn(`collectCycleData: Certificate validation failed for cycle ${cycle.counter} from ${senderInfo} in ${cycle.mode} mode`);
                    archiverLogging_1.ArchiverLogging.logDataSync({
                        sourceArchiver: senderInfo,
                        targetArchiver: Config_1.config.ARCHIVER_IP,
                        cycle: cycle.counter,
                        dataType: 'CYCLE_RECORD',
                        dataHash: cycle.marker,
                        status: 'ERROR',
                        operationId,
                        metrics: {
                            duration: Date.now() - startTime,
                            dataSize: JSON.stringify(cycle).length,
                        },
                        error: 'Certificate validation failed',
                    });
                    break;
                }
                nestedCounters_1.nestedCountersInstance.countEvent('collectCycleData', 'certificate_validation_success_' + cycle.mode, 1);
                Logger.mainLogger.debug(`collectCycleData: Certificate validation successful for cycle ${cycle.counter}`);
            }
            catch (error) {
                nestedCounters_1.nestedCountersInstance.countEvent('collectCycleData', 'certificate_validation_error_' + cycle.mode, 1);
                Logger.mainLogger.error(`collectCycleData: Error during certificate validation for cycle ${cycle.counter}: ${error}`);
                archiverLogging_1.ArchiverLogging.logDataSync({
                    sourceArchiver: senderInfo,
                    targetArchiver: Config_1.config.ARCHIVER_IP,
                    cycle: cycle.counter,
                    dataType: 'CYCLE_RECORD',
                    dataHash: cycle.marker,
                    status: 'ERROR',
                    operationId,
                    metrics: {
                        duration: Date.now() - startTime,
                        dataSize: JSON.stringify(cycle).length,
                    },
                    error: `Certificate validation error: ${error.message}`,
                });
                break;
            }
        }
        receivedCertSigners = cycle.certificates.map((cert) => cert.sign.owner);
        Logger.mainLogger.debug(`collectCycleData: Received ${receivedCertSigners.length} certificate signers for cycle ${cycle.counter}`);
        delete cycle.certificates;
        if (receivedCycleTracker[cycle.counter]) {
            if (receivedCycleTracker[cycle.counter][cycle.marker]) {
                nestedCounters_1.nestedCountersInstance.countEvent('collectCycleData', 'add_signers_to_existing_marker_' + cycle.mode, 1);
                Logger.mainLogger.debug(`collectCycleData: Adding signers to existing marker for cycle ${cycle.counter}`);
                for (const signer of receivedCertSigners)
                    receivedCycleTracker[cycle.counter][cycle.marker]['certSigners'].add(signer);
            }
            else {
                if (!(0, Cycles_1.validateCycleData)(cycle)) {
                    nestedCounters_1.nestedCountersInstance.countEvent('collectCycleData', 'cycle_data_validation_failed_' + cycle.mode, 1);
                    Logger.mainLogger.warn(`collectCycleData: Cycle data validation failed for cycle ${cycle.counter} with marker ${cycle.marker}`);
                    archiverLogging_1.ArchiverLogging.logDataSync({
                        sourceArchiver: senderInfo,
                        targetArchiver: Config_1.config.ARCHIVER_IP,
                        cycle: cycle.counter,
                        dataType: 'CYCLE_RECORD',
                        dataHash: cycle.marker,
                        status: 'ERROR',
                        operationId,
                        metrics: {
                            duration: Date.now() - startTime,
                            dataSize: JSON.stringify(cycle).length,
                        },
                        error: 'Cycle data validation failed',
                    });
                    continue;
                }
                nestedCounters_1.nestedCountersInstance.countEvent('collectCycleData', 'create_new_marker_entry_' + cycle.mode, 1);
                Logger.mainLogger.debug(`collectCycleData: Creating new marker entry for cycle ${cycle.counter} with marker ${cycle.marker}`);
                receivedCycleTracker[cycle.counter][cycle.marker] = {
                    cycleInfo: cycle,
                    certSigners: new Set(receivedCertSigners),
                };
                Logger.mainLogger.debug('Different Cycle Record received', cycle.counter);
            }
            receivedCycleTracker[cycle.counter]['received']++;
            Logger.mainLogger.debug(`collectCycleData: Cycle ${cycle.counter} received count: ${receivedCycleTracker[cycle.counter]['received']}`);
        }
        else {
            if (!(0, Cycles_1.validateCycleData)(cycle)) {
                nestedCounters_1.nestedCountersInstance.countEvent('collectCycleData', 'cycle_data_validation_failed_' + cycle.mode, 1);
                Logger.mainLogger.warn(`collectCycleData: Cycle data validation failed for cycle ${cycle.counter} with marker ${cycle.marker}`);
                archiverLogging_1.ArchiverLogging.logDataSync({
                    sourceArchiver: senderInfo,
                    targetArchiver: Config_1.config.ARCHIVER_IP,
                    cycle: cycle.counter,
                    dataType: 'CYCLE_RECORD',
                    dataHash: cycle.marker,
                    status: 'ERROR',
                    operationId,
                    metrics: {
                        duration: Date.now() - startTime,
                        dataSize: JSON.stringify(cycle).length,
                    },
                    error: 'Cycle data validation failed',
                });
                continue;
            }
            nestedCounters_1.nestedCountersInstance.countEvent('collectCycleData', 'create_new_cycle_tracker_' + cycle.mode, 1);
            Logger.mainLogger.debug(`collectCycleData: Creating new cycle tracker entry for cycle ${cycle.counter}`);
            receivedCycleTracker[cycle.counter] = {
                [cycle.marker]: {
                    cycleInfo: cycle,
                    certSigners: new Set(receivedCertSigners),
                },
                received: 1,
                saved: false,
            };
        }
        if (Config_1.config.VERBOSE)
            Logger.mainLogger.debug('Cycle received', cycle.counter, receivedCycleTracker[cycle.counter]);
        if (NodeList.activeListByIdSorted.length === 0) {
            nestedCounters_1.nestedCountersInstance.countEvent('collectCycleData', 'no_active_nodes_direct_process_' + cycle.mode, 1);
            Logger.mainLogger.debug(`collectCycleData: No active nodes, processing cycle ${cycle.counter} directly`);
            (0, Cycles_1.processCycles)([receivedCycleTracker[cycle.counter][cycle.marker].cycleInfo]);
            continue;
        }
        const requiredSenders = exports.dataSenders.size ? Math.ceil(exports.dataSenders.size / 2) : 1;
        Logger.mainLogger.debug(`collectCycleData: Cycle ${cycle.counter} requires ${requiredSenders} senders, current count: ${receivedCycleTracker[cycle.counter]['received']}`);
        if (receivedCycleTracker[cycle.counter]['received'] >= requiredSenders) {
            nestedCounters_1.nestedCountersInstance.countEvent('collectCycleData', 'enough_senders_process_' + cycle.mode, 1);
            Logger.mainLogger.debug(`collectCycleData: Cycle ${cycle.counter} has enough senders, processing`);
            let bestScore = 0;
            let bestMarker = '';
            const prevMarker = cycleRecordsCache_1.cachedCycleRecords[0].marker;
            Logger.mainLogger.debug(`collectCycleData: Previous marker for scoring: ${prevMarker}`);
            // find the marker with largest sum of its top 3 cert scores
            const markers = Object.entries(receivedCycleTracker[cycle.counter])
                .filter(([key]) => key !== 'saved' && key !== 'received')
                .map(([, value]) => value);
            Logger.mainLogger.debug(`collectCycleData: Found ${markers.length} different markers for cycle ${cycle.counter}`);
            for (const marker of markers) {
                const scores = [];
                for (const signer of marker['certSigners']) {
                    const score = scoreCert(signer, prevMarker);
                    scores.push(score);
                    Logger.mainLogger.debug(`collectCycleData: Cert from ${signer} scored ${score}`);
                }
                // get sum of top 3 scores: sort scores in desc order, then slice off first 3 elements, and add them
                const sum = scores
                    .sort((a, b) => b - a)
                    .slice(0, 3)
                    .reduce((sum, score) => (sum += score), 0);
                Logger.mainLogger.debug(`collectCycleData: Marker ${marker['cycleInfo'].marker} scored ${sum}`);
                if (sum > bestScore) {
                    bestScore = sum;
                    bestMarker = marker['cycleInfo'].marker;
                    Logger.mainLogger.debug(`collectCycleData: New best marker: ${bestMarker} with score ${bestScore}`);
                }
            }
            Logger.mainLogger.debug(`collectCycleData: Processing cycle ${cycle.counter} with best marker ${bestMarker}, score: ${bestScore}`);
            (0, Cycles_1.processCycles)([receivedCycleTracker[cycle.counter][bestMarker].cycleInfo]);
            receivedCycleTracker[cycle.counter]['saved'] = true;
            nestedCounters_1.nestedCountersInstance.countEvent('collectCycleData', 'cycle_processed_successfully_' + cycle.mode, 1);
            archiverLogging_1.ArchiverLogging.logDataSync({
                sourceArchiver: senderInfo,
                targetArchiver: Config_1.config.ARCHIVER_IP,
                cycle: cycle.counter,
                dataType: 'CYCLE_RECORD',
                dataHash: bestMarker,
                status: 'COMPLETE',
                operationId,
                metrics: {
                    duration: Date.now() - startTime,
                    dataSize: JSON.stringify(receivedCycleTracker[cycle.counter][bestMarker].cycleInfo).length,
                },
            });
        }
    }
    if (Object.keys(receivedCycleTracker).length > maxCyclesInCycleTracker) {
        nestedCounters_1.nestedCountersInstance.countEvent('collectCycleData', 'cleanup_old_cycles', 1);
        Logger.mainLogger.debug(`collectCycleData: Cleaning up old cycles, current count: ${Object.keys(receivedCycleTracker).length}`);
        for (const counter of Object.keys(receivedCycleTracker)) {
            // Clear cycles that are older than last maxCyclesInCycleTracker cycles
            if (parseInt(counter) < (0, Cycles_1.getCurrentCycleCounter)() - maxCyclesInCycleTracker) {
                let totalTimes = receivedCycleTracker[counter]['received'];
                let logCycle = false;
                const markers = Object.entries(receivedCycleTracker[counter])
                    .filter(([key]) => key !== 'saved' && key !== 'received')
                    .map(([, value]) => value);
                // If there is more than one marker for this cycle, output the cycle log
                if (markers.length > 1) {
                    logCycle = true;
                    nestedCounters_1.nestedCountersInstance.countEvent('collectCycleData', 'multiple_markers_for_cycle', 1);
                }
                for (const marker of markers) {
                    Logger.mainLogger.debug('Cycle', counter, marker, 
                    /* eslint-disable security/detect-object-injection */
                    logCycle ? lib_types_2.Utils.safeStringify([...receivedCycleTracker[counter][marker]['certSigners']]) : '', logCycle ? receivedCycleTracker[counter][marker] : '');
                }
                if (logCycle)
                    Logger.mainLogger.debug(`Cycle ${counter} has ${markers.length} different markers!`);
                Logger.mainLogger.debug(`Received ${totalTimes} times for cycle counter ${counter}`);
                // eslint-disable-next-line security/detect-object-injection
                delete receivedCycleTracker[counter];
            }
        }
    }
}
exports.collectCycleData = collectCycleData;
function clearCombinedAccountsData() {
    exports.combineAccountsData = {
        accounts: [],
        receipts: [],
    };
}
exports.clearCombinedAccountsData = clearCombinedAccountsData;
exports.dataSenders = new Map();
exports.emitter = new events_1.EventEmitter();
async function replaceDataSender(publicKey) {
    nestedCounters_1.nestedCountersInstance.countEvent('archiver', 'replace_data_sender');
    if (NodeList.getActiveNodeCount() < 2) {
        Logger.mainLogger.debug('There is only one active node in the network. Unable to replace data sender');
        return;
    }
    Logger.mainLogger.debug(`replaceDataSender: replacing ${publicKey}`);
    if (!exports.socketClients.has(publicKey) || !exports.dataSenders.has(publicKey)) {
        Logger.mainLogger.debug('This data sender is not in the subscribed list! and unsubscribing it', publicKey, exports.socketClients.has(publicKey), exports.dataSenders.has(publicKey));
        unsubscribeDataSender(publicKey);
        return;
    }
    unsubscribeDataSender(publicKey);
    // eslint-disable-next-line security/detect-object-injection
    const node = NodeList.byPublicKey.get(publicKey);
    if (node) {
        const nodeIndex = NodeList.activeListByIdSorted.findIndex((node) => node.publicKey === publicKey);
        if (nodeIndex > -1) {
            const subsetIndex = Math.floor(nodeIndex / exports.currentConsensusRadius);
            const subsetNodesList = subsetNodesMapByConsensusRadius.get(subsetIndex);
            if (!subsetNodesList) {
                Logger.mainLogger.error(`There is no nodes in the index ${subsetIndex} of subsetNodesMapByConsensusRadius!`);
                return;
            }
            subscribeNodeFromThisSubset(subsetNodesList);
        }
    }
}
exports.replaceDataSender = replaceDataSender;
async function subscribeNodeForDataTransfer() {
    if (Config_1.config.experimentalSnapshot) {
        await subscribeConsensorsByConsensusRadius();
    }
    else {
        await StateMetaData.subscribeRandomNodeForDataTransfer();
    }
}
exports.subscribeNodeForDataTransfer = subscribeNodeForDataTransfer;
/**
 * Sets 15s timeout
 * Removes sender from dataSenders on timeout
 * Select a new dataSender
 */
function createContactTimeout(publicKey, msg = '') {
    const CONTACT_TIMEOUT_MS = 10 * 1000; // Change contact timeout to 10s
    if (Config_1.config.VERBOSE)
        Logger.mainLogger.debug('Created contact timeout: ' + CONTACT_TIMEOUT_MS, `for ${publicKey}`);
    nestedCounters_1.nestedCountersInstance.countEvent('archiver', 'contact_timeout_created');
    return setTimeout(() => {
        // Logger.mainLogger.debug('nestedCountersInstance', nestedCountersInstance)
        if (nestedCounters_1.nestedCountersInstance)
            nestedCounters_1.nestedCountersInstance.countEvent('archiver', 'contact_timeout');
        Logger.mainLogger.debug('REPLACING sender due to CONTACT timeout', msg, publicKey);
        replaceDataSender(publicKey);
    }, CONTACT_TIMEOUT_MS);
}
exports.createContactTimeout = createContactTimeout;
function addDataSender(sender) {
    exports.dataSenders.set(sender.nodeInfo.publicKey, sender);
}
exports.addDataSender = addDataSender;
async function syncFromNetworkConfig() {
    try {
        // Define the query function to get the network config from a node
        const queryFn = async (node) => {
            const REQUEST_NETCONFIG_TIMEOUT_SECOND = 3; // 3s timeout
            try {
                const response = await P2P.getJson(`http://${node.ip}:${node.port}/netconfig`, REQUEST_NETCONFIG_TIMEOUT_SECOND);
                return response;
            }
            catch (error) {
                Logger.mainLogger.error(`Error querying node ${node.ip}:${node.port}: ${error}`);
                return null;
            }
        };
        // Define the equality function to compare two responses
        const equalityFn = (responseA, responseB) => {
            return responseA?.config?.sharding?.nodesPerConsensusGroup === responseB?.config?.sharding?.nodesPerConsensusGroup;
        };
        // Get the list of 10 max random active nodes or the first node if no active nodes are available
        const nodes = NodeList.getActiveNodeCount() > 0 ? NodeList.getRandomActiveNodes(10) : [NodeList.getFirstNode()];
        // Use robustQuery to get the consensusRadius from multiple nodes
        const tallyItem = await (0, Utils_1.robustQuery)(nodes, queryFn, equalityFn, 3 // Redundancy (minimum 3 nodes should return the same result to reach consensus)
        );
        if (tallyItem?.value?.config?.stateManager) {
            // Updating the Archiver Config as per the latest Network Config
            const { useNewPOQ: newPOQReceipt, configChangeMaxChangesToKeep, configChangeMaxCyclesToKeep, maxCyclesShardDataToKeep, } = tallyItem.value.config.stateManager;
            // const devPublicKeys = tallyItem.value.config.debug.devPublicKeys
            // const devPublicKey =
            //   devPublicKeys &&
            //   Object.keys(devPublicKeys).length >= 3 &&
            //   Object.keys(devPublicKeys).find((key) => devPublicKeys[key] === 3)
            // if (
            //   devPublicKey &&
            //   typeof devPublicKey === typeof config.DevPublicKey &&
            //   devPublicKey !== config.DevPublicKey
            // )
            //   updateConfig({ DevPublicKey: devPublicKey })count query)
            if (!Utils.isUndefined(newPOQReceipt) &&
                typeof newPOQReceipt === typeof Config_1.config.newPOQReceipt &&
                newPOQReceipt !== Config_1.config.newPOQReceipt)
                (0, Config_1.updateConfig)({ newPOQReceipt });
            if (!Utils.isUndefined(configChangeMaxChangesToKeep) &&
                typeof configChangeMaxChangesToKeep === typeof Config_1.config.configChangeMaxChangesToKeep &&
                configChangeMaxChangesToKeep !== Config_1.config.configChangeMaxChangesToKeep)
                (0, Config_1.updateConfig)({ configChangeMaxChangesToKeep });
            if (!Utils.isUndefined(configChangeMaxCyclesToKeep) &&
                typeof configChangeMaxCyclesToKeep === typeof Config_1.config.configChangeMaxCyclesToKeep &&
                configChangeMaxCyclesToKeep !== Config_1.config.configChangeMaxCyclesToKeep)
                (0, Config_1.updateConfig)({ configChangeMaxCyclesToKeep });
            if (!Utils.isUndefined(maxCyclesShardDataToKeep) &&
                typeof maxCyclesShardDataToKeep === typeof Config_1.config.maxCyclesShardDataToKeep &&
                maxCyclesShardDataToKeep !== Config_1.config.maxCyclesShardDataToKeep)
                (0, Config_1.updateConfig)({ maxCyclesShardDataToKeep });
            return tallyItem;
        }
        return null;
    }
    catch (error) {
        Logger.mainLogger.error('❌ Error in syncFromNetworkConfig: ', error);
        return null;
    }
}
async function getConsensusRadius() {
    // If there is no node, return existing currentConsensusRadius
    if (NodeList.isEmpty())
        return exports.currentConsensusRadius;
    const tallyItem = await syncFromNetworkConfig();
    if (tallyItem?.value?.config) {
        const nodesPerEdgeFromConfig = tallyItem.value.config.sharding?.nodesPerEdge;
        const nodesPerConsensusGroupFromConfig = tallyItem.value.config.sharding?.nodesPerConsensusGroup;
        if (!Number.isInteger(nodesPerConsensusGroupFromConfig) || nodesPerConsensusGroupFromConfig <= 0) {
            Logger.mainLogger.error('nodesPerConsensusGroup is not a valid number:', nodesPerConsensusGroupFromConfig);
            return exports.currentConsensusRadius;
        }
        if (!Number.isInteger(nodesPerEdgeFromConfig) || nodesPerEdgeFromConfig <= 0) {
            Logger.mainLogger.error('nodesPerEdge is not a valid number:', nodesPerEdgeFromConfig);
            return exports.currentConsensusRadius;
        }
        if (exports.nodesPerConsensusGroup === nodesPerConsensusGroupFromConfig && exports.nodesPerEdge === nodesPerEdgeFromConfig)
            return exports.currentConsensusRadius;
        exports.nodesPerConsensusGroup = nodesPerConsensusGroupFromConfig;
        exports.nodesPerEdge = nodesPerEdgeFromConfig;
        // Upgrading consensus size to an odd number
        if (exports.nodesPerConsensusGroup % 2 === 0)
            exports.nodesPerConsensusGroup++;
        const consensusRadius = Math.floor((exports.nodesPerConsensusGroup - 1) / 2);
        // Validation: Ensure consensusRadius is a number and greater than zero
        if (typeof consensusRadius !== 'number' || isNaN(consensusRadius) || consensusRadius <= 0) {
            Logger.mainLogger.error('Invalid consensusRadius:', consensusRadius);
            return exports.currentConsensusRadius; // Return the existing currentConsensusRadius in case of invalid consensusRadius
        }
        Logger.mainLogger.debug('consensusRadius', consensusRadius, 'nodesPerConsensusGroup', exports.nodesPerConsensusGroup, 'nodesPerEdge', exports.nodesPerEdge);
        return consensusRadius;
    }
    Logger.mainLogger.error('Failed to get consensusRadius from the network');
    // If no consensus was reached, return the existing currentConsensusRadius
    return exports.currentConsensusRadius;
}
exports.getConsensusRadius = getConsensusRadius;
async function createDataTransferConnection(newSenderInfo) {
    // // Verify node before subscribing for data transfer
    // const status = await verifyNode(newSenderInfo)
    // if (!status) return false
    // Subscribe this node for dataRequest
    const response = await sendDataRequest(newSenderInfo, DataRequestTypes.SUBSCRIBE);
    if (response) {
        initSocketClient(newSenderInfo);
        // Add new dataSender to dataSenders
        const newSender = {
            nodeInfo: newSenderInfo,
            types: [lib_types_1.P2P.SnapshotTypes.TypeNames.CYCLE, lib_types_1.P2P.SnapshotTypes.TypeNames.STATE_METADATA],
            contactTimeout: createContactTimeout(newSenderInfo.publicKey, 'This timeout is created during newSender selection'),
        };
        addDataSender(newSender);
        Logger.mainLogger.debug(`added new sender ${newSenderInfo.publicKey} to dataSenders`);
    }
    return response;
}
exports.createDataTransferConnection = createDataTransferConnection;
async function createNodesGroupByConsensusRadius() {
    const consensusRadius = await getConsensusRadius();
    if (consensusRadius === 0) {
        Logger.mainLogger.error('Consensus radius is 0, unable to create nodes group.');
        return; // Early return to prevent further execution
    }
    exports.currentConsensusRadius = consensusRadius;
    const activeList = [...NodeList.activeListByIdSorted];
    if (Config_1.config.VERBOSE)
        Logger.mainLogger.debug('activeList', activeList.length, activeList);
    let totalNumberOfNodesToSubscribe = Math.ceil(activeList.length / consensusRadius);
    // Only if there are less than 4 activeArchivers and if the consensusRadius is greater than 5
    if (Config_1.config.subscribeToMoreConsensors && State.activeArchivers.length < 4 && exports.currentConsensusRadius > 5) {
        totalNumberOfNodesToSubscribe += totalNumberOfNodesToSubscribe * Config_1.config.extraConsensorsToSubscribe;
    }
    Logger.mainLogger.debug('totalNumberOfNodesToSubscribe', totalNumberOfNodesToSubscribe);
    subsetNodesMapByConsensusRadius = new Map();
    let round = 0;
    for (let i = 0; i < activeList.length; i += consensusRadius) {
        const subsetList = activeList.slice(i, i + consensusRadius);
        subsetNodesMapByConsensusRadius.set(round, subsetList);
        round++;
    }
    if (Config_1.config.VERBOSE)
        Logger.mainLogger.debug('subsetNodesMapByConsensusRadius', subsetNodesMapByConsensusRadius);
}
exports.createNodesGroupByConsensusRadius = createNodesGroupByConsensusRadius;
async function subscribeConsensorsByConsensusRadius() {
    await createNodesGroupByConsensusRadius();
    for (const [i, subsetList] of subsetNodesMapByConsensusRadius) {
        if (Config_1.config.VERBOSE)
            Logger.mainLogger.debug('Round', i, 'subsetList', subsetList, exports.dataSenders.keys());
        subscribeNodeFromThisSubset(subsetList);
    }
}
exports.subscribeConsensorsByConsensusRadius = subscribeConsensorsByConsensusRadius;
async function subscribeNodeFromThisSubset(nodeList) {
    // First check if there is any subscribed node from this subset
    const subscribedNodesFromThisSubset = [];
    for (const node of nodeList) {
        if (exports.dataSenders.has(node.publicKey)) {
            if (Config_1.config.VERBOSE)
                Logger.mainLogger.debug('This node from the subset is in the subscribed list!', node.publicKey);
            subscribedNodesFromThisSubset.push(node.publicKey);
        }
    }
    let numberOfNodesToSubsribe = 1;
    // Only if there are less than 4 activeArchivers and if the consensusRadius is greater than 5
    if (Config_1.config.subscribeToMoreConsensors && State.activeArchivers.length < 4 && exports.currentConsensusRadius > 5) {
        numberOfNodesToSubsribe += Config_1.config.extraConsensorsToSubscribe;
    }
    if (subscribedNodesFromThisSubset.length > numberOfNodesToSubsribe) {
        // If there is more than one subscribed node from this subset, unsubscribe the extra ones
        for (const publicKey of subscribedNodesFromThisSubset.splice(numberOfNodesToSubsribe)) {
            Logger.mainLogger.debug('Unsubscribing extra node from this subset', publicKey);
            unsubscribeDataSender(publicKey);
        }
    }
    if (Config_1.config.VERBOSE)
        Logger.mainLogger.debug('Subscribed nodes from this subset', subscribedNodesFromThisSubset);
    if (subscribedNodesFromThisSubset.length === numberOfNodesToSubsribe)
        return;
    Logger.mainLogger.debug('Subscribing node from this subset!');
    // Pick a new dataSender from this subset
    let subsetList = [...nodeList];
    // Pick a random dataSender
    let newSenderInfo = nodeList[Math.floor(Math.random() * nodeList.length)];
    let connectionStatus = false;
    let retry = 0;
    const MAX_RETRY_SUBSCRIPTION = 3 * numberOfNodesToSubsribe;
    while (retry < MAX_RETRY_SUBSCRIPTION && subscribedNodesFromThisSubset.length < numberOfNodesToSubsribe) {
        if (!exports.dataSenders.has(newSenderInfo.publicKey)) {
            connectionStatus = await createDataTransferConnection(newSenderInfo);
            if (connectionStatus) {
                // Check if the newSender is in the subscribed nodes of this subset
                if (!subscribedNodesFromThisSubset.includes(newSenderInfo.publicKey))
                    subscribedNodesFromThisSubset.push(newSenderInfo.publicKey);
            }
        }
        else {
            // Add the newSender to the subscribed nodes of this subset
            if (!subscribedNodesFromThisSubset.includes(newSenderInfo.publicKey))
                subscribedNodesFromThisSubset.push(newSenderInfo.publicKey);
        }
        subsetList = subsetList.filter((node) => node.publicKey !== newSenderInfo.publicKey);
        if (subsetList.length > 0) {
            newSenderInfo = subsetList[Math.floor(Math.random() * subsetList.length)];
        }
        else {
            subsetList = [...nodeList];
            retry++;
        }
    }
}
exports.subscribeNodeFromThisSubset = subscribeNodeFromThisSubset;
// This function is used for both subscribe and unsubscribe for data request
async function sendDataRequest(nodeInfo, dataRequestType) {
    const dataRequest = {
        dataRequestCycle: (0, Cycles_1.getCurrentCycleCounter)(),
        dataRequestType,
        publicKey: State.getNodeInfo().publicKey,
        nodeInfo: State.getNodeInfo(),
    };
    const taggedDataRequest = Crypto.tag(dataRequest, nodeInfo.publicKey);
    Logger.mainLogger.info(`Sending ${dataRequestType} data request to consensor.`, nodeInfo.ip + ':' + nodeInfo.port);
    let reply = false;
    const REQUEST_DATA_TIMEOUT_SECOND = 2; // 2s timeout
    const response = await P2P.postJson(`http://${nodeInfo.ip}:${nodeInfo.port}/requestdata`, taggedDataRequest, REQUEST_DATA_TIMEOUT_SECOND);
    Logger.mainLogger.debug('/requestdata response', response, nodeInfo.ip + ':' + nodeInfo.port);
    if (response && response.success)
        reply = response.success;
    return reply;
}
exports.sendDataRequest = sendDataRequest;
const clearDataSenders = async () => {
    for (const [publicKey] of exports.dataSenders) {
        unsubscribeDataSender(publicKey);
    }
    await Utils.sleep(2000); // Wait for 2s to make sure all dataSenders are unsubscribed
    exports.dataSenders.clear();
    exports.socketClients.clear();
    subsetNodesMapByConsensusRadius.clear();
};
exports.clearDataSenders = clearDataSenders;
function calcIncomingTimes(record) {
    const SECOND = 1000;
    const cycleDuration = record.duration * SECOND;
    const quarterDuration = cycleDuration / 4;
    const start = record.start * SECOND + cycleDuration;
    const startQ1 = start;
    const startQ2 = start + quarterDuration;
    const startQ3 = start + 2 * quarterDuration;
    const startQ4 = start + 3 * quarterDuration;
    const end = start + cycleDuration;
    return { quarterDuration, startQ1, startQ2, startQ3, startQ4, end };
}
exports.calcIncomingTimes = calcIncomingTimes;
async function joinNetwork(nodeList, isFirstTime) {
    Logger.mainLogger.debug('Is firstTime', isFirstTime);
    if (!isFirstTime) {
        const isJoined = await checkJoinStatus(nodeList);
        if (isJoined) {
            return isJoined;
        }
    }
    Logger.mainLogger.debug('nodeList To Submit Join Request', nodeList);
    // try to get latestCycleRecord with a robust query
    const latestCycle = await (0, Cycles_1.getNewestCycleFromConsensors)(nodeList);
    // Figure out when Q1 is from the latestCycle
    const { startQ1 } = calcIncomingTimes(latestCycle);
    const shuffledNodes = [...nodeList];
    Utils.shuffleArray(shuffledNodes);
    // Wait until a Q1 then send join request to active nodes
    let untilQ1 = startQ1 - Date.now();
    while (untilQ1 < 0) {
        untilQ1 += latestCycle.duration * 1000;
    }
    Logger.mainLogger.debug(`Waiting ${untilQ1 + 500} ms for Q1 before sending join...`);
    await Utils.sleep(untilQ1 + 500); // Not too early
    // Create a fresh join request, so that the request timestamp range is acceptable
    const request = P2P.createArchiverJoinRequest();
    await submitJoin(nodeList, request);
    // Wait approx. one cycle then check again
    Logger.mainLogger.debug('Waiting approx. one cycle then checking again...');
    await Utils.sleep(latestCycle.duration * 1000 + 500);
    return false;
}
exports.joinNetwork = joinNetwork;
async function submitJoin(nodes, joinRequest) {
    // Send the join request to a handful of the active node all at once:w
    const selectedNodes = Utils.getRandom(nodes, Math.min(nodes.length, 5));
    Logger.mainLogger.debug(`Sending join request to ${selectedNodes.map((n) => `${n.ip}:${n.port}`)}`);
    for (const node of selectedNodes) {
        const response = await P2P.postJson(`http://${node.ip}:${node.port}/joinarchiver`, joinRequest);
        Logger.mainLogger.debug('Join request response:', response);
    }
}
exports.submitJoin = submitJoin;
async function sendLeaveRequest(nodes) {
    const leaveRequest = P2P.createArchiverLeaveRequest();
    Logger.mainLogger.debug(`Sending leave request to ${nodes.map((n) => `${n.ip}:${n.port}`)}`);
    const promises = nodes.map((node) => (0, customHttpFunctions_1.customFetch)(`http://${node.ip}:${node.port}/leavingarchivers`, {
        method: 'post',
        body: lib_types_2.Utils.safeStringify(leaveRequest),
        headers: { 'Content-Type': 'application/json' },
        timeout: 2 * 1000, // 2s timeout
    }).then((res) => res.json()));
    await Promise.allSettled(promises)
        .then((responses) => {
        let i = 0;
        let isLeaveRequestSent = false;
        for (const response of responses) {
            // eslint-disable-next-line security/detect-object-injection
            const node = nodes[i];
            if (response.status === 'fulfilled') {
                const res = response.value;
                if (res.success)
                    isLeaveRequestSent = true;
                Logger.mainLogger.debug(`Leave request response from ${node.ip}:${node.port}:`, res);
            }
            else
                Logger.mainLogger.debug(`Node is not responding ${node.ip}:${node.port}`);
            i++;
        }
        Logger.mainLogger.debug('isLeaveRequestSent', isLeaveRequestSent);
    })
        .catch((error) => {
        // Handle any errors that occurred
        console.error(error);
    });
}
exports.sendLeaveRequest = sendLeaveRequest;
async function sendActiveRequest() {
    Logger.mainLogger.debug('Sending Active Request to the network!');
    const latestCycleInfo = await CycleDB.queryLatestCycleRecords(1);
    const latestCycle = latestCycleInfo[0];
    // Figure out when Q1 is from the latestCycle
    const { startQ1 } = calcIncomingTimes(latestCycle);
    // Wait until a Q1 then send active request to active nodes
    let untilQ1 = startQ1 - Date.now();
    while (untilQ1 < 0) {
        untilQ1 += latestCycle.duration * 1000;
    }
    Logger.mainLogger.debug(`Waiting ${untilQ1 + 500} ms for Q1 before sending active...`);
    await Utils.sleep(untilQ1 + 500); // Not too early
    const activeRequest = P2P.createArchiverActiveRequest();
    // Send the active request to a handful of the active node all at once:w
    const nodes = NodeList.getRandomActiveNodes(5);
    Logger.mainLogger.debug(`Sending active request to ${nodes.map((n) => `${n.ip}:${n.port}`)}`);
    const promises = nodes.map((node) => (0, customHttpFunctions_1.customFetch)(`http://${node.ip}:${node.port}/activearchiver`, {
        method: 'post',
        body: lib_types_2.Utils.safeStringify(activeRequest),
        headers: { 'Content-Type': 'application/json' },
        timeout: 2 * 1000, // 2s timeout
    }).then((res) => res.json()));
    await Promise.allSettled(promises)
        .then((responses) => {
        let i = 0;
        for (const response of responses) {
            // eslint-disable-next-line security/detect-object-injection
            const node = nodes[i];
            if (response.status === 'fulfilled') {
                const res = response.value;
                Logger.mainLogger.debug(`Active request response from ${node.ip}:${node.port}:`, res);
            }
            else
                Logger.mainLogger.debug(`Node is not responding ${node.ip}:${node.port}`);
            i++;
        }
    })
        .catch((error) => {
        // Handle any errors that occurred
        console.error(error);
    });
    // Wait approx. one cycle then check again
    Logger.mainLogger.debug('Waiting approx. one cycle then checking again...');
    await Utils.sleep(latestCycle.duration * 1000 + 500);
}
exports.sendActiveRequest = sendActiveRequest;
async function getCycleDuration() {
    const response = (await (0, API_1.queryFromArchivers)(API_1.RequestDataType.CYCLE, { count: 1 }));
    if (response && response.cycleInfo) {
        return response.cycleInfo[0].duration;
    }
    return 0;
}
exports.getCycleDuration = getCycleDuration;
/*
  checkJoinStatus checks if the current archiver node is joined to a network.
  This queries by the /joinedArchiver endpoint on the nodes and returns joining status based on majority response.
*/
async function checkJoinStatus(activeNodes) {
    Logger.mainLogger.debug('checkJoinStatus: Checking join status');
    const ourNodeInfo = State.getNodeInfo();
    const queryFn = async (node) => {
        const url = `http://${node.ip}:${node.port}/joinedArchiver/${ourNodeInfo.publicKey}`;
        try {
            return (await (0, P2P_1.getJson)(url));
        }
        catch (e) {
            Logger.mainLogger.error(`Error querying node ${node.ip}:${node.port}: ${e}`);
            throw e;
        }
    };
    try {
        const joinStatus = await (0, Utils_1.robustQuery)(activeNodes, queryFn);
        Logger.mainLogger.debug(`checkJoinStatus: Join status: ${joinStatus.value.isJoined}`);
        return joinStatus.value.isJoined;
    }
    catch (e) {
        Logger.mainLogger.error(`Error in checkJoinStatus: ${e}`);
        return false;
    }
}
exports.checkJoinStatus = checkJoinStatus;
// This will be used once activeArchivers field is added to the cycle record
async function checkActiveStatus() {
    Logger.mainLogger.debug('Checking active status');
    const ourNodeInfo = State.getNodeInfo();
    try {
        const latestCycle = await (0, Cycles_1.getNewestCycleFromArchivers)();
        if (latestCycle && latestCycle['activeArchivers']) {
            const activeArchivers = latestCycle['activeArchivers'];
            Logger.mainLogger.debug('cycle counter', latestCycle.counter);
            Logger.mainLogger.debug('Active archivers', activeArchivers);
            const isActive = activeArchivers.some((a) => a.publicKey === ourNodeInfo.publicKey);
            Logger.mainLogger.debug('isActive', isActive);
            return isActive;
        }
        else {
            return false;
        }
    }
    catch (e) {
        Logger.mainLogger.error(e);
        return false;
    }
}
exports.checkActiveStatus = checkActiveStatus;
async function getTotalDataFromArchivers() {
    const res = (await (0, API_1.queryFromArchivers)(API_1.RequestDataType.TOTALDATA, {}, QUERY_TIMEOUT_MAX));
    // @ts-ignore
    if (!res || (res.success !== undefined && res.success === false)) {
        return null;
    }
    return res;
}
exports.getTotalDataFromArchivers = getTotalDataFromArchivers;
async function syncGenesisAccountsFromArchiver() {
    let complete = false;
    let startAccount = 0;
    let endAccount = startAccount + MAX_ACCOUNTS_PER_REQUEST;
    let totalGenesisAccounts = 0;
    // const totalExistingGenesisAccounts =
    //   await AccountDB.queryAccountCountBetweenCycles(0, 5);
    // if (totalExistingGenesisAccounts > 0) {
    //   // Let's assume it has synced data for now, update to sync account count between them
    //   return;
    // }
    const res = (await (0, API_1.queryFromArchivers)(API_1.RequestDataType.ACCOUNT, { startCycle: GENESIS_ACCOUNTS_CYCLE_RANGE.startCycle, endCycle: GENESIS_ACCOUNTS_CYCLE_RANGE.endCycle }, QUERY_TIMEOUT_MAX));
    if (Config_1.config.VERBOSE)
        Logger.mainLogger.error('Genesis Total Accounts Response', lib_types_2.Utils.safeStringify(res));
    if (res && (res.totalAccounts || res.totalAccounts === 0)) {
        totalGenesisAccounts = res.totalAccounts;
        Logger.mainLogger.debug('TotalGenesis Accounts', totalGenesisAccounts);
    }
    else {
        Logger.mainLogger.error('Genesis Total Accounts Query', 'Invalid download response');
        return;
    }
    if (totalGenesisAccounts <= 0)
        return;
    let page = 1;
    while (!complete) {
        Logger.mainLogger.debug(`Downloading accounts from ${startAccount} to ${endAccount}`);
        const response = (await (0, API_1.queryFromArchivers)(API_1.RequestDataType.ACCOUNT, {
            startCycle: GENESIS_ACCOUNTS_CYCLE_RANGE.startCycle,
            endCycle: GENESIS_ACCOUNTS_CYCLE_RANGE.endCycle,
            page,
        }, QUERY_TIMEOUT_MAX));
        if (response && response.accounts) {
            if (response.accounts.length < MAX_ACCOUNTS_PER_REQUEST) {
                complete = true;
                Logger.mainLogger.debug('Download completed for accounts');
            }
            Logger.mainLogger.debug(`Downloaded accounts`, response.accounts.length);
            await (0, Collector_1.storeAccountData)({ accounts: response.accounts });
            startAccount = endAccount + 1;
            endAccount += MAX_ACCOUNTS_PER_REQUEST;
            page++;
        }
        else {
            Logger.mainLogger.debug('Genesis Accounts Query', 'Invalid download response');
        }
        // await sleep(1000);
    }
    Logger.mainLogger.debug('Sync genesis accounts completed!');
}
exports.syncGenesisAccountsFromArchiver = syncGenesisAccountsFromArchiver;
async function syncGenesisTransactionsFromArchiver() {
    let complete = false;
    let startTransaction = 0;
    let endTransaction = startTransaction + MAX_ACCOUNTS_PER_REQUEST; // Sames as number of accounts per request
    let totalGenesisTransactions = 0;
    const res = (await (0, API_1.queryFromArchivers)(API_1.RequestDataType.TRANSACTION, {
        startCycle: GENESIS_ACCOUNTS_CYCLE_RANGE.startCycle,
        endCycle: GENESIS_ACCOUNTS_CYCLE_RANGE.endCycle,
    }, QUERY_TIMEOUT_MAX));
    if (Config_1.config.VERBOSE)
        Logger.mainLogger.error('Genesis Total Transaction Response', lib_types_2.Utils.safeStringify(res));
    if (res && (res.totalTransactions || res.totalTransactions === 0)) {
        totalGenesisTransactions = res.totalTransactions;
        Logger.mainLogger.debug('TotalGenesis Transactions', totalGenesisTransactions);
    }
    else {
        Logger.mainLogger.error('Genesis Total Transaction Query', 'Invalid download response');
        return;
    }
    if (totalGenesisTransactions <= 0)
        return;
    let page = 1;
    while (!complete) {
        Logger.mainLogger.debug(`Downloading transactions from ${startTransaction} to ${endTransaction}`);
        const response = (await (0, API_1.queryFromArchivers)(API_1.RequestDataType.TRANSACTION, {
            startCycle: GENESIS_ACCOUNTS_CYCLE_RANGE.startCycle,
            endCycle: GENESIS_ACCOUNTS_CYCLE_RANGE.endCycle,
            page,
        }, QUERY_TIMEOUT_MAX));
        if (response && response.transactions) {
            if (response.transactions.length < MAX_ACCOUNTS_PER_REQUEST) {
                complete = true;
                Logger.mainLogger.debug('Download completed for transactions');
            }
            Logger.mainLogger.debug(`Downloaded transactions`, response.transactions.length);
            await (0, Collector_1.storeAccountData)({ receipts: response.transactions });
            startTransaction = endTransaction + 1;
            endTransaction += MAX_ACCOUNTS_PER_REQUEST;
            page++;
        }
        else {
            Logger.mainLogger.debug('Genesis Transactions Query', 'Invalid download response');
        }
        // await sleep(1000);
    }
    Logger.mainLogger.debug('Sync genesis transactions completed!');
}
exports.syncGenesisTransactionsFromArchiver = syncGenesisTransactionsFromArchiver;
async function syncGenesisAccountsFromConsensor(totalGenesisAccounts = 0, firstConsensor) {
    if (totalGenesisAccounts <= 0)
        return;
    let startAccount = 0;
    // let combineAccountsData = [];
    let totalDownloadedAccounts = 0;
    while (startAccount <= totalGenesisAccounts) {
        Logger.mainLogger.debug(`Downloading accounts from ${startAccount}`);
        const response = (await P2P.getJson(`http://${firstConsensor.ip}:${firstConsensor.port}/genesis_accounts?start=${startAccount}`, QUERY_TIMEOUT_MAX));
        if (response && response.accounts) {
            if (response.accounts.length < MAX_ACCOUNTS_PER_REQUEST) {
                Logger.mainLogger.debug('Download completed for accounts');
            }
            Logger.mainLogger.debug(`Downloaded accounts`, response.accounts.length);
            // TODO - update to include receipts data also
            await (0, Collector_1.storeAccountData)({ accounts: response.accounts });
            // combineAccountsData = [...combineAccountsData, ...response.accounts];
            totalDownloadedAccounts += response.accounts.length;
            startAccount += MAX_ACCOUNTS_PER_REQUEST;
        }
        else {
            Logger.mainLogger.debug('Genesis Accounts Query', 'Invalid download response');
        }
        // await sleep(1000);
    }
    Logger.mainLogger.debug(`Total downloaded accounts`, totalDownloadedAccounts);
    // await storeAccountData(combineAccountsData);
    Logger.mainLogger.debug('Sync genesis accounts completed!');
}
exports.syncGenesisAccountsFromConsensor = syncGenesisAccountsFromConsensor;
async function buildNodeListFromStoredCycle(lastStoredCycle) {
    Logger.mainLogger.debug('lastStoredCycle', lastStoredCycle);
    Logger.mainLogger.debug(`Syncing till cycle ${lastStoredCycle.counter}...`);
    const cyclesToGet = 2 * Math.floor(Math.sqrt(lastStoredCycle.active)) + 2;
    Logger.mainLogger.debug(`Cycles to get is ${cyclesToGet}`);
    const CycleChain = [];
    const squasher = new CycleParser_1.ChangeSquasher();
    CycleChain.unshift(lastStoredCycle);
    squasher.addChange((0, CycleParser_1.parse)(CycleChain[0]));
    do {
        // Get prevCycles from the network
        let end = CycleChain[0].counter - 1;
        let start = end - cyclesToGet;
        if (start < 0)
            start = 0;
        if (end < start)
            end = start;
        Logger.mainLogger.debug(`Getting cycles ${start} - ${end}...`);
        const prevCycles = await CycleDB.queryCycleRecordsBetween(start, end);
        // If prevCycles is empty, start over
        if (prevCycles.length < 1)
            throw new Error('Got empty previous cycles');
        prevCycles.sort((a, b) => (a.counter > b.counter ? -1 : 1));
        // Add prevCycles to our cycle chain
        let prepended = 0;
        for (const prevCycle of prevCycles) {
            // Prepend the cycle to our cycle chain
            CycleChain.unshift(prevCycle);
            squasher.addChange((0, CycleParser_1.parse)(prevCycle));
            prepended++;
            if (squasher.final.updated.length >= (0, CycleParser_1.activeNodeCount)(lastStoredCycle) &&
                squasher.final.added.length >= (0, CycleParser_1.totalNodeCount)(lastStoredCycle)) {
                break;
            }
        }
        Logger.mainLogger.debug(`Got ${squasher.final.updated.length} active nodes, need ${(0, CycleParser_1.activeNodeCount)(lastStoredCycle)}`);
        Logger.mainLogger.debug(`Got ${squasher.final.added.length} total nodes, need ${(0, CycleParser_1.totalNodeCount)(lastStoredCycle)}`);
        if (squasher.final.added.length < (0, CycleParser_1.totalNodeCount)(lastStoredCycle))
            Logger.mainLogger.debug('Short on nodes. Need to get more cycles. Cycle:' + lastStoredCycle.counter);
        // If you weren't able to prepend any of the prevCycles, start over
        if (prepended < 1)
            throw new Error('Unable to prepend any previous cycles');
    } while (squasher.final.updated.length < (0, CycleParser_1.activeNodeCount)(lastStoredCycle) ||
        squasher.final.added.length < (0, CycleParser_1.totalNodeCount)(lastStoredCycle));
    (0, CycleParser_1.applyNodeListChange)(squasher.final);
    Logger.mainLogger.debug('NodeList after sync', NodeList.getActiveList());
    Cycles.setCurrentCycleCounter(lastStoredCycle.counter);
    Cycles.setCurrentCycleMarker(lastStoredCycle.marker);
    Cycles.setCurrentCycleDuration(lastStoredCycle.duration);
    Logger.mainLogger.debug('Latest cycle after sync', lastStoredCycle.counter);
}
exports.buildNodeListFromStoredCycle = buildNodeListFromStoredCycle;
async function syncCyclesAndNodeList(lastStoredCycleCount = 0) {
    // Get the networks newest cycle as the anchor point for sync
    Logger.mainLogger.debug('Getting newest cycle...');
    const cycleToSyncTo = await (0, Cycles_1.getNewestCycleFromArchivers)();
    Logger.mainLogger.debug('cycleToSyncTo', cycleToSyncTo);
    Logger.mainLogger.debug(`Syncing till cycle ${cycleToSyncTo.counter}...`);
    const cyclesToGet = 2 * Math.floor(Math.sqrt(cycleToSyncTo.active)) + 2;
    Logger.mainLogger.debug(`Cycles to get is ${cyclesToGet}`);
    const CycleChain = [];
    const squasher = new CycleParser_1.ChangeSquasher();
    CycleChain.unshift(cycleToSyncTo);
    squasher.addChange((0, CycleParser_1.parse)(CycleChain[0]));
    do {
        // Get prevCycles from the network
        let end = CycleChain[0].counter - 1;
        let start = end - cyclesToGet;
        if (start < 0)
            start = 0;
        if (end < start)
            end = start;
        Logger.mainLogger.debug(`Getting cycles ${start} - ${end}...`);
        const prevCycles = await (0, Cycles_1.fetchCycleRecords)(start, end);
        // If prevCycles is empty, start over
        if (prevCycles.length < 1)
            throw new Error('Got empty previous cycles');
        prevCycles.sort((a, b) => (a.counter > b.counter ? -1 : 1));
        // Add prevCycles to our cycle chain
        let prepended = 0;
        for (const prevCycle of prevCycles) {
            // Stop prepending prevCycles if one of them is invalid
            if ((0, Cycles_1.validateCycle)(prevCycle, CycleChain[0]) === false) {
                Logger.mainLogger.error(`Record ${prevCycle.counter} failed validation`);
                break;
            }
            // Prepend the cycle to our cycle chain
            CycleChain.unshift(prevCycle);
            squasher.addChange((0, CycleParser_1.parse)(prevCycle));
            prepended++;
            if (squasher.final.updated.length >= (0, CycleParser_1.activeNodeCount)(cycleToSyncTo) &&
                squasher.final.added.length >= (0, CycleParser_1.totalNodeCount)(cycleToSyncTo)) {
                break;
            }
        }
        Logger.mainLogger.debug(`Got ${squasher.final.updated.length} active nodes, need ${(0, CycleParser_1.activeNodeCount)(cycleToSyncTo)}`);
        Logger.mainLogger.debug(`Got ${squasher.final.added.length} total nodes, need ${(0, CycleParser_1.totalNodeCount)(cycleToSyncTo)}`);
        if (squasher.final.added.length < (0, CycleParser_1.totalNodeCount)(cycleToSyncTo))
            Logger.mainLogger.debug('Short on nodes. Need to get more cycles. Cycle:' + cycleToSyncTo.counter);
        // If you weren't able to prepend any of the prevCycles, start over
        if (prepended < 1)
            throw new Error('Unable to prepend any previous cycles');
    } while (squasher.final.updated.length < (0, CycleParser_1.activeNodeCount)(cycleToSyncTo) ||
        squasher.final.added.length < (0, CycleParser_1.totalNodeCount)(cycleToSyncTo));
    (0, CycleParser_1.applyNodeListChange)(squasher.final);
    Logger.mainLogger.debug('NodeList after sync', NodeList.getActiveList());
    for (let i = 0; i < CycleChain.length; i++) {
        // eslint-disable-next-line security/detect-object-injection
        const record = CycleChain[i];
        Cycles.CycleChain.set(record.counter, { ...record });
        if (i === CycleChain.length - 1)
            await (0, Collector_1.storeCycleData)(CycleChain);
        Cycles.setCurrentCycleCounter(record.counter);
        Cycles.setCurrentCycleMarker(record.marker);
    }
    Logger.mainLogger.debug('Cycle chain is synced. Size of CycleChain', Cycles.CycleChain.size);
    // Download old cycle Records
    let endCycle = CycleChain[0].counter - 1;
    Logger.mainLogger.debug('endCycle counter', endCycle, 'lastStoredCycleCount', lastStoredCycleCount);
    if (endCycle > lastStoredCycleCount) {
        Logger.mainLogger.debug(`Downloading old cycles from cycles ${lastStoredCycleCount} to cycle ${endCycle}!`);
    }
    let savedCycleRecord = CycleChain[0];
    while (endCycle > lastStoredCycleCount) {
        let nextEnd = endCycle - MAX_CYCLES_PER_REQUEST;
        if (nextEnd < 0)
            nextEnd = 0;
        Logger.mainLogger.debug(`Getting cycles ${nextEnd} - ${endCycle} ...`);
        const prevCycles = await (0, Cycles_1.fetchCycleRecords)(nextEnd, endCycle);
        // If prevCycles is empty, start over
        if (!prevCycles || prevCycles.length < 1)
            throw new Error('Got empty previous cycles');
        prevCycles.sort((a, b) => (a.counter > b.counter ? -1 : 1));
        // Add prevCycles to our cycle chain
        const combineCycles = [];
        for (const prevCycle of prevCycles) {
            // Stop saving prevCycles if one of them is invalid
            if ((0, Cycles_1.validateCycle)(prevCycle, savedCycleRecord) === false) {
                Logger.mainLogger.error(`Record ${prevCycle.counter} failed validation`);
                Logger.mainLogger.debug('fail', prevCycle, savedCycleRecord);
                break;
            }
            savedCycleRecord = prevCycle;
            combineCycles.push(prevCycle);
        }
        await (0, Collector_1.storeCycleData)(combineCycles);
        endCycle = nextEnd - 1;
    }
}
exports.syncCyclesAndNodeList = syncCyclesAndNodeList;
async function syncCyclesAndNodeListV2(activeArchivers, lastStoredCycleCount = 0) {
    // Sync validator list and get the latest cycle from the network
    Logger.mainLogger.debug('Syncing validators and latest cycle...');
    const syncResult = await (0, sync_v2_1.syncV2)(activeArchivers);
    let cycleToSyncTo;
    if (syncResult.isOk()) {
        cycleToSyncTo = syncResult.value;
    }
    else {
        throw syncResult.error;
    }
    Logger.mainLogger.debug('cycleToSyncTo', cycleToSyncTo);
    Logger.mainLogger.debug(`Syncing till cycle ${cycleToSyncTo.counter}...`);
    exports.currentConsensusRadius = await getConsensusRadius();
    await (0, Cycles_1.processCycles)([cycleToSyncTo]);
    // Download old cycle Records
    await downloadOldCycles(cycleToSyncTo, lastStoredCycleCount);
    return true;
}
exports.syncCyclesAndNodeListV2 = syncCyclesAndNodeListV2;
async function syncCyclesBetweenCycles(lastStoredCycle = 0, cycleToSyncTo = 0) {
    const MAX_RETRIES = 3;
    let retryCount = 0;
    let startCycle = lastStoredCycle;
    let endCycle = startCycle + MAX_CYCLES_PER_REQUEST;
    while (cycleToSyncTo > startCycle) {
        if (endCycle > cycleToSyncTo)
            endCycle = cycleToSyncTo;
        Logger.mainLogger.debug(`Downloading cycles from ${startCycle} to ${endCycle}`);
        let success = false;
        retryCount = 0;
        while (!success && retryCount < MAX_RETRIES) {
            const res = (await (0, API_1.queryFromArchivers)(API_1.RequestDataType.CYCLE, {
                start: startCycle,
                end: endCycle,
            }, QUERY_TIMEOUT_MAX));
            if (res && res.cycleInfo) {
                const cycles = res.cycleInfo;
                Logger.mainLogger.debug(`Downloaded cycles`, cycles.length);
                let validCyclesCount = 0;
                for (const cycle of cycles) {
                    if (!(0, Cycles_1.validateCycleData)(cycle)) {
                        Logger.mainLogger.debug('Found invalid cycle data');
                        continue;
                    }
                    await (0, Cycles_1.processCycles)([cycle]);
                    validCyclesCount++;
                }
                success = true;
                if (cycles.length < MAX_CYCLES_PER_REQUEST || validCyclesCount === 0) {
                    startCycle += Math.max(cycles.length, 1);
                    endCycle = startCycle + MAX_CYCLES_PER_REQUEST;
                    if (startCycle >= cycleToSyncTo) {
                        Logger.mainLogger.debug('Sync cycles completed!');
                        return true;
                    }
                    break;
                }
            }
            else {
                Logger.mainLogger.debug(`Invalid cycle download response, attempt ${retryCount + 1} of ${MAX_RETRIES}`);
                retryCount++;
                if (retryCount >= MAX_RETRIES) {
                    Logger.mainLogger.error('Max retries reached for cycle download');
                    return false;
                }
            }
        }
        if (success) {
            startCycle = endCycle + 1;
            endCycle += MAX_CYCLES_PER_REQUEST;
        }
    }
    return true;
}
exports.syncCyclesBetweenCycles = syncCyclesBetweenCycles;
async function syncReceipts() {
    const MAX_RETRIES = 3;
    let retryCount = 0;
    let response = await getTotalDataFromArchivers();
    if (!response || response.totalReceipts < 0) {
        return;
    }
    let { totalReceipts } = response;
    if (totalReceipts < 1)
        return;
    let complete = false;
    let start = 0;
    let end = start + MAX_RECEIPTS_PER_REQUEST;
    while (!complete) {
        if (end >= totalReceipts) {
            response = await getTotalDataFromArchivers();
            if (response && response.totalReceipts > 0) {
                if (response.totalReceipts > totalReceipts)
                    totalReceipts = response.totalReceipts;
                Logger.mainLogger.debug('totalReceiptsToSync', totalReceipts);
            }
        }
        Logger.mainLogger.debug(`Downloading receipts from ${start} to ${end}`);
        let success = false;
        retryCount = 0;
        while (!success && retryCount < MAX_RETRIES) {
            const res = (await (0, API_1.queryFromArchivers)(API_1.RequestDataType.RECEIPT, {
                start: start,
                end: end,
            }, QUERY_TIMEOUT_MAX));
            if (res && res.receipts) {
                const downloadedReceipts = res.receipts;
                Logger.mainLogger.debug(`Downloaded receipts`, downloadedReceipts.length);
                await (0, Collector_1.storeReceiptData)(downloadedReceipts);
                success = true;
                if (downloadedReceipts.length < MAX_RECEIPTS_PER_REQUEST) {
                    start += downloadedReceipts.length;
                    end = start + MAX_RECEIPTS_PER_REQUEST;
                    response = await getTotalDataFromArchivers();
                    if (response && response.totalReceipts > 0) {
                        if (response.totalReceipts > totalReceipts)
                            totalReceipts = response.totalReceipts;
                        if (start >= totalReceipts) {
                            complete = true;
                            Logger.mainLogger.debug('Download receipts completed');
                        }
                    }
                }
            }
            else {
                Logger.mainLogger.debug(`Invalid download response, attempt ${retryCount + 1} of ${MAX_RETRIES}`);
                retryCount++;
                if (retryCount >= MAX_RETRIES) {
                    Logger.mainLogger.error('Max retries reached for receipt download');
                    start = end + 1;
                    end += MAX_RECEIPTS_PER_REQUEST;
                    // Check if we've passed total receipts after incrementing
                    if (start >= totalReceipts) {
                        complete = true;
                    }
                }
            }
        }
        if (success) {
            start = end + 1;
            end += MAX_RECEIPTS_PER_REQUEST;
        }
    }
    Logger.mainLogger.debug('Sync receipts data completed!');
}
exports.syncReceipts = syncReceipts;
async function syncReceiptsByCycle(lastStoredReceiptCycle = 0, cycleToSyncTo = 0) {
    let totalCycles = cycleToSyncTo;
    let totalReceipts = 0;
    if (cycleToSyncTo === 0) {
        const response = await getTotalDataFromArchivers();
        if (!response || response.totalReceipts < 0) {
            return false;
        }
        totalCycles = response.totalCycles;
        totalReceipts = response.totalReceipts;
    }
    let startCycle = lastStoredReceiptCycle;
    let endCycle = startCycle + MAX_BETWEEN_CYCLES_PER_REQUEST;
    let receiptsCountToSyncBetweenCycles = 0;
    let savedReceiptsCountBetweenCycles = 0;
    let totalSavedReceiptsCount = 0;
    let retryCount = 0;
    const MAX_RETRIES = 3;
    while (true) {
        if (endCycle > totalCycles) {
            endCycle = totalCycles;
            totalSavedReceiptsCount = await ReceiptDB.queryReceiptCount();
        }
        if (cycleToSyncTo > 0) {
            if (startCycle > cycleToSyncTo) {
                Logger.mainLogger.debug(`Sync receipts data completed!`);
                return true;
            }
        }
        else {
            if (totalSavedReceiptsCount >= totalReceipts) {
                const res = await getTotalDataFromArchivers();
                if (res && res.totalReceipts > 0) {
                    if (res.totalReceipts > totalReceipts)
                        totalReceipts = res.totalReceipts;
                    if (res.totalCycles > totalCycles)
                        totalCycles = res.totalCycles;
                    Logger.mainLogger.debug('totalReceiptsToSync', totalReceipts, 'totalSavedReceipts', totalSavedReceiptsCount);
                    if (totalSavedReceiptsCount === totalReceipts) {
                        Logger.mainLogger.debug('Sync receipts data completed!');
                        return true;
                    }
                }
            }
        }
        if (startCycle > endCycle) {
            Logger.mainLogger.error(`Got some issues in syncing receipts. Receipts query startCycle ${startCycle} is greater than endCycle ${endCycle}`);
            return false;
        }
        Logger.mainLogger.debug(`Downloading receipts from cycle ${startCycle} to cycle ${endCycle}`);
        let response = (await (0, API_1.queryFromArchivers)(API_1.RequestDataType.RECEIPT, {
            startCycle,
            endCycle,
            type: 'count',
        }, QUERY_TIMEOUT_MAX));
        if (response && response.receipts > 0) {
            receiptsCountToSyncBetweenCycles = response.receipts;
            let page = 1;
            savedReceiptsCountBetweenCycles = 0;
            while (savedReceiptsCountBetweenCycles < receiptsCountToSyncBetweenCycles) {
                const res = (await (0, API_1.queryFromArchivers)(API_1.RequestDataType.RECEIPT, {
                    startCycle,
                    endCycle,
                    page,
                }, QUERY_TIMEOUT_MAX));
                if (res && res.receipts) {
                    const downloadedReceipts = res.receipts;
                    Logger.mainLogger.debug(`Downloaded receipts`, downloadedReceipts.length);
                    await (0, Collector_1.storeReceiptData)(downloadedReceipts);
                    savedReceiptsCountBetweenCycles += downloadedReceipts.length;
                    if (savedReceiptsCountBetweenCycles > receiptsCountToSyncBetweenCycles) {
                        response = (await (0, API_1.queryFromArchivers)(API_1.RequestDataType.RECEIPT, {
                            startCycle,
                            endCycle,
                            type: 'count',
                        }, QUERY_TIMEOUT_MAX));
                        if (response && response.receipts)
                            receiptsCountToSyncBetweenCycles = response.receipts;
                        if (receiptsCountToSyncBetweenCycles > savedReceiptsCountBetweenCycles) {
                            savedReceiptsCountBetweenCycles -= downloadedReceipts.length;
                            continue;
                        }
                    }
                    Logger.mainLogger.debug('savedReceiptsCountBetweenCycles', savedReceiptsCountBetweenCycles, 'receiptsCountToSyncBetweenCycles', receiptsCountToSyncBetweenCycles);
                    if (savedReceiptsCountBetweenCycles > receiptsCountToSyncBetweenCycles) {
                        Logger.mainLogger.debug(`It has downloaded more receipts than it has in cycles between ${startCycle} and ${endCycle} !`);
                    }
                    totalSavedReceiptsCount += downloadedReceipts.length;
                    page++;
                    retryCount = 0;
                }
                else {
                    Logger.mainLogger.debug('Invalid download response');
                    retryCount++;
                    if (retryCount >= MAX_RETRIES) {
                        Logger.mainLogger.error('Max retries reached for invalid download response');
                        return false;
                    }
                    continue;
                }
            }
            Logger.mainLogger.debug(`Download receipts completed for ${startCycle} - ${endCycle}`);
            // Update checkpoint status for completed cycles
            startCycle = endCycle + 1;
            endCycle += MAX_BETWEEN_CYCLES_PER_REQUEST;
            retryCount = 0;
        }
        else {
            receiptsCountToSyncBetweenCycles = response.receipts;
            if (receiptsCountToSyncBetweenCycles === 0) {
                startCycle = endCycle + 1;
                endCycle += MAX_BETWEEN_CYCLES_PER_REQUEST;
                retryCount = 0;
                continue;
            }
            Logger.mainLogger.debug('Invalid download response');
            retryCount++;
            if (retryCount >= MAX_RETRIES) {
                Logger.mainLogger.error('Max retries reached for invalid download response');
                return false;
            }
        }
    }
}
exports.syncReceiptsByCycle = syncReceiptsByCycle;
const syncOriginalTxs = async () => {
    const MAX_RETRIES = 3;
    let retryCount = 0;
    let response = await getTotalDataFromArchivers();
    if (!response || response.totalOriginalTxs < 0) {
        return;
    }
    let { totalOriginalTxs } = response;
    if (totalOriginalTxs < 1)
        return;
    let complete = false;
    let start = 0;
    let end = start + MAX_ORIGINAL_TXS_PER_REQUEST;
    while (!complete) {
        if (end >= totalOriginalTxs) {
            // If the number of new original txs to sync is within MAX_ORIGINAL_TXS_PER_REQUEST => Update to the latest totalOriginalTxs.
            response = await getTotalDataFromArchivers();
            if (response && response.totalOriginalTxs > 0) {
                if (response.totalOriginalTxs > totalOriginalTxs)
                    totalOriginalTxs = response.totalOriginalTxs;
                Logger.mainLogger.debug('totalOriginalTxs: ', totalOriginalTxs);
            }
        }
        Logger.mainLogger.debug(`Downloading Original-Txs from ${start} to ${end}`);
        let success = false;
        retryCount = 0;
        while (!success && retryCount < MAX_RETRIES) {
            const res = await (0, API_1.queryFromArchivers)(API_1.RequestDataType.ORIGINALTX, {
                start: start,
                end: end,
            }, QUERY_TIMEOUT_MAX);
            if (res && res.originalTxs) {
                const downloadedOriginalTxs = res.originalTxs;
                Logger.mainLogger.debug('Downloaded Original-Txs: ', downloadedOriginalTxs.length);
                await (0, Collector_1.storeOriginalTxData)(downloadedOriginalTxs);
                success = true;
                if (downloadedOriginalTxs.length < MAX_ORIGINAL_TXS_PER_REQUEST) {
                    start += downloadedOriginalTxs.length;
                    end = start + MAX_ORIGINAL_TXS_PER_REQUEST;
                    response = await getTotalDataFromArchivers();
                    if (response && response.totalOriginalTxs > 0) {
                        if (response.totalOriginalTxs > totalOriginalTxs)
                            totalOriginalTxs = response.totalOriginalTxs;
                        if (start >= totalOriginalTxs) {
                            complete = true;
                            Logger.mainLogger.debug('Download Original-Txs Completed!');
                        }
                    }
                    break;
                }
            }
            else {
                Logger.mainLogger.debug(`Invalid Original-Tx download response, attempt ${retryCount + 1} of ${MAX_RETRIES}`);
                retryCount++;
                if (retryCount >= MAX_RETRIES) {
                    Logger.mainLogger.error('Max retries reached for Original-Tx download');
                    start = end + 1;
                    end += MAX_ORIGINAL_TXS_PER_REQUEST;
                    if (start >= totalOriginalTxs) {
                        complete = true;
                    }
                }
            }
        }
        if (success) {
            start = end + 1;
            end += MAX_ORIGINAL_TXS_PER_REQUEST;
        }
    }
    Logger.mainLogger.debug('Sync Original-Txs Data Completed!');
};
exports.syncOriginalTxs = syncOriginalTxs;
const syncOriginalTxsByCycle = async (lastStoredOriginalTxCycle = 0, cycleToSyncTo = 0) => {
    let totalCycles = cycleToSyncTo;
    let totalOriginalTxs = 0;
    if (cycleToSyncTo === 0) {
        const response = await getTotalDataFromArchivers();
        if (!response || response.totalOriginalTxs < 1) {
            return;
        }
        totalCycles = response.totalCycles;
        totalOriginalTxs = response.totalOriginalTxs;
    }
    const complete = false;
    let startCycle = lastStoredOriginalTxCycle;
    let endCycle = startCycle + MAX_BETWEEN_CYCLES_PER_REQUEST;
    let originalTxCountToSyncBetweenCycles = 0;
    let savedOriginalTxCountBetweenCycles = 0;
    let totalSavedOriginalTxCount = 0;
    while (!complete) {
        if (endCycle > totalCycles) {
            endCycle = totalCycles;
            totalSavedOriginalTxCount = await OriginalTxDB.queryOriginalTxDataCount();
        }
        if (cycleToSyncTo > 0) {
            if (startCycle > cycleToSyncTo) {
                Logger.mainLogger.debug(`Sync originalTXs data completed!`);
                break;
            }
        }
        else {
            if (totalSavedOriginalTxCount >= totalOriginalTxs) {
                const res = await getTotalDataFromArchivers();
                if (res && res.totalOriginalTxs > 0) {
                    if (res.totalOriginalTxs > totalOriginalTxs)
                        totalOriginalTxs = res.totalOriginalTxs;
                    if (res.totalCycles > totalCycles)
                        totalCycles = res.totalCycles;
                    Logger.mainLogger.debug('totalOriginalTxsToSync: ', totalOriginalTxs, 'totalSavedOriginalTxs: ', totalSavedOriginalTxCount);
                    if (totalSavedOriginalTxCount === totalOriginalTxs) {
                        Logger.mainLogger.debug('Sync Original-Tx data completed!');
                        break;
                    }
                }
            }
        }
        if (startCycle > endCycle) {
            Logger.mainLogger.error(`Got some issues in syncing Original-Tx data. Original-Tx query startCycle ${startCycle} is greater than endCycle ${endCycle}`);
            break;
        }
        Logger.mainLogger.debug(`Downloading Original-Tx data from cycle ${startCycle} to cycle ${endCycle}`);
        let response = (await (0, API_1.queryFromArchivers)(API_1.RequestDataType.ORIGINALTX, {
            startCycle,
            endCycle,
            type: 'count',
        }, QUERY_TIMEOUT_MAX));
        if (response && response.originalTxs > 0) {
            originalTxCountToSyncBetweenCycles = response.originalTxs;
            let page = 1;
            savedOriginalTxCountBetweenCycles = 0;
            while (savedOriginalTxCountBetweenCycles < originalTxCountToSyncBetweenCycles) {
                const res = (await (0, API_1.queryFromArchivers)(API_1.RequestDataType.ORIGINALTX, {
                    startCycle,
                    endCycle,
                    page,
                }, QUERY_TIMEOUT_MAX));
                if (res && res.originalTxs) {
                    const downloadedOriginalTxs = res.originalTxs;
                    Logger.mainLogger.debug('Downloaded Original-Txs: ', downloadedOriginalTxs.length);
                    await (0, Collector_1.storeOriginalTxData)(downloadedOriginalTxs);
                    savedOriginalTxCountBetweenCycles += downloadedOriginalTxs.length;
                    if (savedOriginalTxCountBetweenCycles > originalTxCountToSyncBetweenCycles) {
                        response = (await (0, API_1.queryFromArchivers)(API_1.RequestDataType.ORIGINALTX, {
                            startCycle,
                            endCycle,
                            type: 'count',
                        }, QUERY_TIMEOUT_MAX));
                        if (response && response.originalTxs)
                            originalTxCountToSyncBetweenCycles = response.originalTxs;
                        if (originalTxCountToSyncBetweenCycles > savedOriginalTxCountBetweenCycles) {
                            savedOriginalTxCountBetweenCycles -= downloadedOriginalTxs.length;
                            continue;
                        }
                    }
                    Logger.mainLogger.debug('savedOriginalTxCountBetweenCycles', savedOriginalTxCountBetweenCycles, 'originalTxCountToSyncBetweenCycles', originalTxCountToSyncBetweenCycles);
                    if (savedOriginalTxCountBetweenCycles > originalTxCountToSyncBetweenCycles) {
                        Logger.mainLogger.debug(`It has downloaded more originalTxsData than it has in cycles between ${startCycle} and ${endCycle} !`);
                    }
                    totalSavedOriginalTxCount += downloadedOriginalTxs.length;
                    page++;
                }
                else {
                    Logger.mainLogger.debug('Invalid Original-Txs download response');
                    continue;
                }
            }
            Logger.mainLogger.debug(`Download Original-Txs completed for ${startCycle} - ${endCycle}`);
            startCycle = endCycle + 1;
            endCycle += MAX_BETWEEN_CYCLES_PER_REQUEST;
        }
        else {
            originalTxCountToSyncBetweenCycles = response.originalTxs;
            if (originalTxCountToSyncBetweenCycles === 0) {
                startCycle = endCycle + 1;
                endCycle += MAX_BETWEEN_CYCLES_PER_REQUEST;
                continue;
            }
            Logger.mainLogger.debug('Invalid Original-Txs download response');
            continue;
        }
    }
};
exports.syncOriginalTxsByCycle = syncOriginalTxsByCycle;
const syncCyclesAndTxsData = async (lastStoredCycleCount = 0, lastStoredReceiptCount = 0, lastStoredOriginalTxCount = 0) => {
    const MAX_RETRIES = 3;
    let retryCount = 0;
    let response = await getTotalDataFromArchivers();
    if (!response || response.totalCycles < 0 || response.totalReceipts < 0) {
        return;
    }
    const { totalCycles, totalReceipts } = response;
    Logger.mainLogger.debug('totalCycles', totalCycles, 'lastStoredCycleCount', lastStoredCycleCount);
    Logger.mainLogger.debug('totalReceipts', totalReceipts, 'lastStoredReceiptCount', lastStoredReceiptCount);
    // Logger.mainLogger.debug(
    //   'totalOriginalTxs',
    //   totalOriginalTxs,
    //   'lastStoredOriginalTxCount',
    //   lastStoredOriginalTxCount
    // )
    if (totalCycles === lastStoredCycleCount &&
        totalReceipts === lastStoredReceiptCount
    // && totalOriginalTxs === lastStoredOriginalTxCount
    ) {
        Logger.mainLogger.debug('The archiver has synced the lastest cycle ,receipts and originalTxs data!');
        return;
    }
    let totalReceiptsToSync = totalReceipts;
    // let totalOriginalTxsToSync = totalOriginalTxs
    let totalCyclesToSync = totalCycles;
    let completeForReceipt = false;
    // let completeForOriginalTx = false
    let completeForCycle = false;
    let startReceipt = lastStoredReceiptCount;
    // let startOriginalTx = lastStoredOriginalTxCount
    let startCycle = lastStoredCycleCount;
    let endReceipt = startReceipt + MAX_RECEIPTS_PER_REQUEST;
    // let endOriginalTx = startOriginalTx + MAX_ORIGINAL_TXS_PER_REQUEST
    let endCycle = startCycle + MAX_CYCLES_PER_REQUEST;
    if (totalCycles === lastStoredCycleCount)
        completeForCycle = true;
    if (totalReceipts === lastStoredReceiptCount)
        completeForReceipt = true;
    // if (totalOriginalTxs === lastStoredOriginalTxCount) completeForOriginalTx = true
    while (!completeForReceipt ||
        !completeForCycle
    //  || !completeForOriginalTx
    ) {
        if (endReceipt >= totalReceiptsToSync ||
            endCycle >= totalCyclesToSync
        // || endOriginalTx >= totalOriginalTxsToSync
        ) {
            response = await getTotalDataFromArchivers();
            if (response &&
                response.totalReceipts &&
                response.totalCycles
            // && response.totalOriginalTxs
            ) {
                if (response.totalReceipts !== totalReceiptsToSync) {
                    completeForReceipt = false;
                    totalReceiptsToSync = response.totalReceipts;
                }
                // if (response.totalOriginalTxs !== totalOriginalTxsToSync) {
                //   completeForOriginalTx = false
                //   totalOriginalTxsToSync = response.totalOriginalTxs
                // }
                if (response.totalCycles !== totalCyclesToSync) {
                    completeForCycle = false;
                    totalCyclesToSync = response.totalCycles;
                }
                if (totalReceiptsToSync < startReceipt) {
                    completeForReceipt = true;
                }
                // if (totalOriginalTxsToSync < startOriginalTx) {
                //   completeForOriginalTx = true
                // }
                if (totalCyclesToSync < startCycle) {
                    completeForCycle = true;
                }
                Logger.mainLogger.debug('totalReceiptsToSync', totalReceiptsToSync, 
                // 'totalOriginalTxsToSync',
                // totalOriginalTxsToSync,
                'totalCyclesToSync', totalCyclesToSync);
            }
        }
        if (!completeForReceipt) {
            Logger.mainLogger.debug(`Downloading receipts from ${startReceipt} to ${endReceipt}`);
            let success = false;
            retryCount = 0;
            while (!success && retryCount < MAX_RETRIES) {
                const res = (await (0, API_1.queryFromArchivers)(API_1.RequestDataType.RECEIPT, {
                    start: startReceipt,
                    end: endReceipt,
                }, QUERY_TIMEOUT_MAX));
                if (res && res.receipts) {
                    const downloadedReceipts = res.receipts;
                    Logger.mainLogger.debug(`Downloaded receipts`, downloadedReceipts.length);
                    await (0, Collector_1.storeReceiptData)(downloadedReceipts);
                    success = true;
                    if (downloadedReceipts.length < MAX_ORIGINAL_TXS_PER_REQUEST) {
                        startReceipt += downloadedReceipts.length + 1;
                        endReceipt += downloadedReceipts.length + MAX_ORIGINAL_TXS_PER_REQUEST;
                    }
                }
                else {
                    Logger.mainLogger.debug(`Invalid download response, attempt ${retryCount + 1} of ${MAX_RETRIES}`);
                    retryCount++;
                    if (retryCount >= MAX_RETRIES) {
                        Logger.mainLogger.error('Max retries reached for receipt download');
                    }
                }
            }
            if (success) {
                startReceipt = endReceipt + 1;
                endReceipt += MAX_ORIGINAL_TXS_PER_REQUEST;
            }
        }
        // if (!completeForOriginalTx) {
        //   Logger.mainLogger.debug(`Downloading Original-Txs from ${startOriginalTx} to ${endOriginalTx}`)
        //   let success = false
        //   retryCount = 0
        //   while (!success && retryCount < MAX_RETRIES) {
        //     const res = (await queryFromArchivers(
        //       RequestDataType.ORIGINALTX,
        //       {
        //         start: startOriginalTx,
        //         end: endOriginalTx,
        //       },
        //       QUERY_TIMEOUT_MAX
        //     )) as ArchiverOriginalTxResponse
        //     if (res && res.originalTxs) {
        //       const downloadedOriginalTxs = res.originalTxs as OriginalTxDB.OriginalTxData[]
        //       Logger.mainLogger.debug(`Downloaded Original-Txs: `, downloadedOriginalTxs.length)
        //       await storeOriginalTxData(downloadedOriginalTxs)
        //       success = true
        //       if (downloadedOriginalTxs.length < MAX_ORIGINAL_TXS_PER_REQUEST) {
        //         startOriginalTx += downloadedOriginalTxs.length + 1
        //         endOriginalTx += downloadedOriginalTxs.length + MAX_ORIGINAL_TXS_PER_REQUEST
        //         break
        //       }
        //     } else {
        //       Logger.mainLogger.debug(`Invalid Original-Tx download response, attempt ${retryCount + 1} of ${MAX_RETRIES}`)
        //       retryCount++
        //       if (retryCount >= MAX_RETRIES) {
        //         Logger.mainLogger.error('Max retries reached for Original-Tx download')
        //         startOriginalTx = endOriginalTx + 1
        //         endOriginalTx += MAX_ORIGINAL_TXS_PER_REQUEST
        //       }
        //       continue
        //     }
        //   }
        //   if (success) {
        //     startOriginalTx = endOriginalTx + 1
        //     endOriginalTx += MAX_ORIGINAL_TXS_PER_REQUEST
        //   }
        // }
        if (!completeForCycle) {
            Logger.mainLogger.debug(`Downloading cycles from ${startCycle} to ${endCycle}`);
            let success = false;
            retryCount = 0;
            while (!success && retryCount < MAX_RETRIES) {
                const res = (await (0, API_1.queryFromArchivers)(API_1.RequestDataType.CYCLE, {
                    start: startCycle,
                    end: endCycle,
                }, QUERY_TIMEOUT_MAX));
                if (res && res.cycleInfo) {
                    const cycles = res.cycleInfo;
                    Logger.mainLogger.debug(`Downloaded cycles`, cycles.length);
                    for (const cycle of cycles) {
                        if (!(0, Cycles_1.validateCycleData)(cycle)) {
                            Logger.mainLogger.debug('Found invalid cycle data');
                            continue;
                        }
                        await (0, Cycles_1.processCycles)([cycle]);
                    }
                    success = true;
                    if (cycles.length < MAX_CYCLES_PER_REQUEST) {
                        startCycle += cycles.length + 1;
                        endCycle += cycles.length + MAX_CYCLES_PER_REQUEST;
                    }
                }
                else {
                    Logger.mainLogger.debug(`Invalid cycle download response, attempt ${retryCount + 1} of ${MAX_RETRIES}`);
                    retryCount++;
                    if (retryCount >= MAX_RETRIES) {
                        Logger.mainLogger.error('Max retries reached for cycle download');
                    }
                }
            }
            if (success) {
                startCycle = endCycle + 1;
                endCycle += MAX_CYCLES_PER_REQUEST;
            }
        }
    }
    Logger.mainLogger.debug('Sync Cycle, Receipt & Original-Tx data completed!');
};
exports.syncCyclesAndTxsData = syncCyclesAndTxsData;
const syncCyclesAndTxsDataBetweenCycles = async (lastStoredCycle = 0, cycleToSyncTo = 0) => {
    Logger.mainLogger.debug(`Syncing cycles and txs data between cycles ${lastStoredCycle} and ${cycleToSyncTo}`);
    await syncCyclesBetweenCycles(lastStoredCycle, cycleToSyncTo);
    await syncReceiptsByCycle(lastStoredCycle, cycleToSyncTo);
    // await syncOriginalTxsByCycle(lastStoredCycle, cycleToSyncTo)
};
exports.syncCyclesAndTxsDataBetweenCycles = syncCyclesAndTxsDataBetweenCycles;
// // simple method to validate old data; it's not good when there are multiple archivers, the receipts saving order may not be the same
// export async function compareWithOldReceiptsData(
//   archiver: State.ArchiverNodeInfo,
//   lastReceiptCount = 0
// ) {
//   let downloadedReceipts
//   const response: any = await P2P.getJson(
//     `http://${archiver.ip}:${archiver.port}/receipt?start=${
//       lastReceiptCount - 10 > 0 ? lastReceiptCount - 10 : 0
//     }&end=${lastReceiptCount}`
//   )
//   if (response && response.receipts) {
//     downloadedReceipts = response.receipts
//   } else {
//     throw Error(
//       `Can't fetch data from receipt ${
//         lastReceiptCount - 10 > 0 ? lastReceiptCount - 10 : 0
//       } to receipt ${lastReceiptCount}  from archiver ${archiver}`
//     )
//   }
//   let oldReceipts = await ReceiptDB.queryReceipts(
//     lastReceiptCount - 10 > 0 ? lastReceiptCount - 10 : 0,
//     lastReceiptCount
//   )
//   // downloadedReceipts.sort((a, b) =>
//   //   a.cycleRecord.counter > b.cycleRecord.counter ? 1 : -1
//   // );
//   // oldReceipts.sort((a, b) =>
//   //   a.cycleRecord.counter > b.cycleRecord.counter ? 1 : -1
//   // );
//   let success = false
//   let receiptsToMatchCount = 10
//   for (let i = 0; i < downloadedReceipts.length; i++) {
//     let downloadedReceipt = downloadedReceipts[i]
//     const oldReceipt = oldReceipts[i]
//     if (oldReceipt.counter) delete oldReceipt.counter
//     console.log(downloadedReceipt.receiptId, oldReceipt.receiptId)
//     if (downloadedReceipt.receiptId !== oldReceipt.receiptId) {
//       return {
//         success,
//         receiptsToMatchCount,
//       }
//     }
//     success = true
//     receiptsToMatchCount--
//   }
//   return { success, receiptsToMatchCount }
// }
async function compareWithOldOriginalTxsData(lastStoredOriginalTxCycle = 0) {
    const numberOfCyclesTocompare = 10;
    let success = false;
    let matchedCycle = 0;
    const endCycle = lastStoredOriginalTxCycle;
    const startCycle = endCycle - numberOfCyclesTocompare > 0 ? endCycle - numberOfCyclesTocompare : 0;
    const response = (await (0, API_1.queryFromArchivers)(API_1.RequestDataType.ORIGINALTX, {
        startCycle,
        endCycle,
        type: 'tally',
    }, QUERY_TIMEOUT_MAX));
    if (!response || !response.originalTxs) {
        Logger.mainLogger.error(`Can't fetch original tx data from cycle ${startCycle} to cycle ${endCycle} from archivers`);
        return { success, matchedCycle };
    }
    const downloadedOriginalTxsByCycles = response.originalTxs;
    const oldOriginalTxCountByCycle = await OriginalTxDB.queryOriginalTxDataCountByCycles(startCycle, endCycle);
    for (let i = 0; i < downloadedOriginalTxsByCycles.length; i++) {
        // eslint-disable-next-line security/detect-object-injection
        const downloadedOriginalTx = downloadedOriginalTxsByCycles[i];
        // eslint-disable-next-line security/detect-object-injection
        const oldOriginalTx = oldOriginalTxCountByCycle[i];
        Logger.mainLogger.debug(downloadedOriginalTx, oldOriginalTx);
        if (!downloadedOriginalTx ||
            !oldOriginalTx ||
            downloadedOriginalTx.cycle !== oldOriginalTx.cycle ||
            downloadedOriginalTx.originalTxDataCount !== oldOriginalTx.originalTxDataCount) {
            return {
                success,
                matchedCycle,
            };
        }
        success = true;
        matchedCycle = downloadedOriginalTx.cycle;
    }
    success = true;
    return { success, matchedCycle };
}
exports.compareWithOldOriginalTxsData = compareWithOldOriginalTxsData;
async function compareWithOldReceiptsData(lastStoredReceiptCycle = 0) {
    const numberOfCyclesTocompare = 10;
    let success = false;
    let matchedCycle = 0;
    const endCycle = lastStoredReceiptCycle;
    const startCycle = endCycle - numberOfCyclesTocompare > 0 ? endCycle - numberOfCyclesTocompare : 0;
    const response = (await (0, API_1.queryFromArchivers)(API_1.RequestDataType.RECEIPT, {
        startCycle,
        endCycle,
        type: 'tally',
    }, QUERY_TIMEOUT_MAX));
    if (!response || !response.receipts) {
        Logger.mainLogger.error(`Can't fetch receipts data from cycle ${startCycle} to cycle ${endCycle}  from archivers`);
        return { success, matchedCycle };
    }
    const downloadedReceiptCountByCycles = response.receipts;
    const oldReceiptCountByCycle = await ReceiptDB.queryReceiptCountByCycles(startCycle, endCycle);
    for (let i = 0; i < downloadedReceiptCountByCycles.length; i++) {
        // eslint-disable-next-line security/detect-object-injection
        const downloadedReceipt = downloadedReceiptCountByCycles[i];
        // eslint-disable-next-line security/detect-object-injection
        const oldReceipt = oldReceiptCountByCycle[i];
        Logger.mainLogger.debug(downloadedReceipt, oldReceipt);
        if (!downloadedReceipt ||
            !oldReceipt ||
            downloadedReceipt.cycle !== oldReceipt.cycle ||
            downloadedReceipt.receiptCount !== oldReceipt.receiptCount) {
            return {
                success,
                matchedCycle,
            };
        }
        success = true;
        matchedCycle = downloadedReceipt.cycle;
    }
    success = true;
    return { success, matchedCycle };
}
exports.compareWithOldReceiptsData = compareWithOldReceiptsData;
async function compareWithOldCyclesData(lastCycleCounter = 0) {
    try {
        const numberOfCyclesTocompare = 10;
        const start = lastCycleCounter - numberOfCyclesTocompare;
        const end = lastCycleCounter;
        const response = (await (0, API_1.queryFromArchivers)(API_1.RequestDataType.CYCLE, {
            start,
            end,
        }, QUERY_TIMEOUT_MAX));
        if (!response && !response.cycleInfo) {
            throw Error(`Can't fetch data from cycle ${start} to cycle ${end}  from archivers`);
        }
        const downloadedCycles = response.cycleInfo;
        const oldCycles = await CycleDB.queryCycleRecordsBetween(start, end);
        let success = false;
        let matchedCycle = 0;
        for (let i = 0; i < downloadedCycles.length; i++) {
            // eslint-disable-next-line security/detect-object-injection
            const downloadedCycle = downloadedCycles[i];
            // eslint-disable-next-line security/detect-object-injection
            const oldCycle = oldCycles[i];
            if (!downloadedCycle ||
                !oldCycle ||
                lib_types_2.Utils.safeStringify(downloadedCycle) !== lib_types_2.Utils.safeStringify(oldCycle)) {
                if (Config_1.config.VERBOSE) {
                    Logger.mainLogger.error('Mismatched cycle Number', downloadedCycle.counter, oldCycle.counter);
                }
                return {
                    success,
                    matchedCycle,
                };
            }
            success = true;
            matchedCycle = downloadedCycle.counter;
        }
        return { success, matchedCycle };
    }
    catch (error) {
        Logger.mainLogger.error('compareWithOldCyclesData error: ' + error);
        return { success: false, matchedCycle: 0 };
    }
}
exports.compareWithOldCyclesData = compareWithOldCyclesData;
async function downloadOldCycles(cycleToSyncTo, lastStoredCycleCount) {
    let endCycle = cycleToSyncTo.counter - 1;
    Logger.mainLogger.debug('endCycle counter', endCycle, 'lastStoredCycleCount', lastStoredCycleCount);
    if (endCycle > lastStoredCycleCount) {
        Logger.mainLogger.debug(`Downloading old cycles from cycles ${lastStoredCycleCount} to cycle ${endCycle}!`);
    }
    let savedCycleRecord = cycleToSyncTo;
    const MAX_RETRY_COUNT = 3;
    let retryCount = 0;
    while (endCycle > lastStoredCycleCount) {
        let startCycle = endCycle - MAX_CYCLES_PER_REQUEST;
        if (startCycle < 0)
            startCycle = 0;
        if (startCycle < lastStoredCycleCount)
            startCycle = lastStoredCycleCount;
        Logger.mainLogger.debug(`Getting cycles ${startCycle} - ${endCycle} ...`);
        const res = (await (0, API_1.queryFromArchivers)(API_1.RequestDataType.CYCLE, {
            start: startCycle,
            end: endCycle,
        }, QUERY_TIMEOUT_MAX));
        if (!res || !res.cycleInfo || !Array.isArray(res.cycleInfo) || res.cycleInfo.length === 0) {
            Logger.mainLogger.error(`Can't fetch data from cycle ${startCycle} to cycle ${endCycle}  from archivers`);
            if (retryCount < MAX_RETRY_COUNT) {
                retryCount++;
                continue;
            }
            else {
                endCycle = startCycle - 1;
                retryCount = 0;
            }
        }
        const prevCycles = res.cycleInfo;
        if (prevCycles)
            prevCycles.sort((a, b) => (a.counter > b.counter ? -1 : 1));
        const combineCycles = [];
        for (const prevCycle of prevCycles) {
            if ((0, Cycles_1.validateCycle)(prevCycle, savedCycleRecord) === false) {
                Logger.mainLogger.error(`Record ${prevCycle.counter} failed validation`);
                Logger.mainLogger.debug('fail', prevCycle, savedCycleRecord);
            }
            savedCycleRecord = prevCycle;
            combineCycles.push(prevCycle);
        }
        await (0, Collector_1.storeCycleData)(combineCycles);
        endCycle = startCycle - 1;
    }
}
/**
 * Syncs cycle data for a specific cycle
 * @param cycle The cycle number to sync
 * @returns True if successful, false otherwise
 */
async function syncCycleData(cycle) {
    const MAX_RETRIES = 3;
    let retryCount = 0;
    let success = false;
    Logger.mainLogger.debug(`syncCycleData: Starting sync for cycle ${cycle}`);
    Logger.mainLogger.debug(`syncCycleData: Active nodes count: ${NodeList.activeListByIdSorted.length}`);
    while (!success && retryCount < MAX_RETRIES) {
        try {
            Logger.mainLogger.debug(`syncCycleData: Attempt ${retryCount + 1} for cycle ${cycle}`);
            const res = (await (0, API_1.queryFromArchivers)(API_1.RequestDataType.CYCLE, {
                start: cycle,
                end: cycle,
            }, QUERY_TIMEOUT_MAX));
            if (res && res.cycleInfo && res.cycleInfo.length > 0) {
                const cycleData = res.cycleInfo[0];
                Logger.mainLogger.debug(`syncCycleData: Received data for cycle ${cycle}, marker: ${cycleData.marker}`);
                if (!(0, Cycles_1.validateCycleData)(cycleData)) {
                    Logger.mainLogger.error(`syncCycleData: Invalid cycle data for cycle ${cycle}`);
                    Logger.mainLogger.error(`syncCycleData: Cycle validation failed, checking marker computation...`);
                    nestedCounters_1.nestedCountersInstance.countEvent('archiver', 'cycle_validation_failed - ' + cycle);
                    // Debug marker computation
                    const cycleDataCopy = { ...cycleData };
                    delete cycleDataCopy.marker;
                    const computedMarker = Cycles.computeCycleMarker(cycleDataCopy);
                    Logger.mainLogger.error(`syncCycleData: Computed marker: ${computedMarker}, received marker: ${cycleData.marker}`);
                    retryCount++;
                    continue;
                }
                await (0, Cycles_1.processCycles)([cycleData]);
                Logger.mainLogger.debug(`syncCycleData: Successfully synced and processed cycle ${cycle}`);
                // Successfully synced cycle data for cycle
                success = true;
                return true;
            }
            else {
                Logger.mainLogger.error(`syncCycleData: Failed to get cycle data for cycle ${cycle}, attempt ${retryCount + 1} of ${MAX_RETRIES}`);
                retryCount++;
            }
        }
        catch (error) {
            Logger.mainLogger.error(`syncCycleData: Error syncing cycle data for cycle ${cycle}: ${error}`);
            retryCount++;
        }
    }
    Logger.mainLogger.error(`syncCycleData: All attempts to sync cycle ${cycle} failed`);
    return false;
}
exports.syncCycleData = syncCycleData;
// We want to check that all of the provided certs are actually valid. This means we need to check that
// they all have the same marker, and that that marker is the same as the one of the original record.
// We also want all of the signer to actually be active node. And of course, the certs must pass sign
// verification. If we have already verified a cert in the past, we can skip it. It is possble for a
// malicious node to use valid certs from a honest record to get this function to return true. However,
// it will also have to make sure the inpMarker is the same as the markers as it is in the certs. If it
// does this, then the validateCycleData() function that gets called later will fail
function validateCerts(certs, certSigners, inpMarker, cycleData) {
    nestedCounters_1.nestedCountersInstance.countEvent('validateCerts', 'validation', 1);
    Logger.mainLogger.debug(`validateCerts: Validating ${certs.length} certificates against marker ${inpMarker}`);
    for (const cert of certs) {
        const cleanCert = {
            marker: cert.marker,
            sign: cert.sign,
        };
        if (cleanCert.marker !== inpMarker) {
            nestedCounters_1.nestedCountersInstance.countEvent('validateCerts', 'markerMismatch', 1);
            validationTracker.add({ cycle: cycleData });
            return false;
        }
        if (NodeList.activeListByIdSorted.some((node) => node.publicKey === cleanCert.sign.owner) === false) {
            nestedCounters_1.nestedCountersInstance.countEvent('validateCerts', 'badOwner', 1);
            Logger.mainLogger.warn(`validateCerts: bad owner ${cleanCert.sign.owner} not found in active nodes`);
            return false;
        }
        if (certSigners.has(cert.sign.owner)) {
            nestedCounters_1.nestedCountersInstance.countEvent('validateCerts', 'skipExistingSigner', 1);
            Logger.mainLogger.debug(`validateCerts: Skipping already verified cert from ${cert.sign.owner}`);
            continue;
        }
        if (!Crypto.verify(cleanCert)) {
            nestedCounters_1.nestedCountersInstance.countEvent('validateCerts', 'badSignature', 1);
            Logger.mainLogger.warn(`validateCerts: bad signature from ${cleanCert.sign.owner}`);
            return false;
        }
        nestedCounters_1.nestedCountersInstance.countEvent('validateCerts', 'validCert', 1);
    }
    Logger.mainLogger.debug(`validateCerts: All certificates validated successfully`);
    return true;
}
function scoreCert(pubKey, prevMarker) {
    try {
        const node = NodeList.byPublicKey.get(pubKey);
        const id = node.id; // get node id from cert pub key
        const obj = { id };
        const hid = Crypto.hashObj(obj); // Omar - use hash of id so the cert is not made by nodes that are near based on node id
        const out = (0, general_1.XOR)(prevMarker, hid);
        // will also nerf if foundationNode is undefined, which is will be for already active nodes when we
        // first turn on the addFoundationNodeAttribute flag under the current implementation
        if (Config_1.config.nerfNonFoundationCertScores && !node.foundationNode) {
            return out & 0x0fffffff;
        }
        return out;
    }
    catch (err) {
        Logger.mainLogger.error('scoreCert ERR:', err);
        return 0;
    }
}
exports.scoreCert = scoreCert;
// this function is needed since the cycle record is changed after Q3/Q4. Thus, the cycle certs will contain
// the marker of the cycle as it existed in Q3/Q4. However, the cycle that we ceived at the start of the
// function has been changed, so its marker has also been changed. If we try to check this new mark against
// the markers inside the certs, the validation will obviously fail. So we want to revert those changes on a
// deep copy so that we can get the original record
function getRecordWithoutPostQ3Changes(cycle) {
    Logger.mainLogger.debug(`getRecordWithoutPostQ3Changes: Processing cycle ${cycle.counter}`);
    const cycleCopy = lib_types_2.Utils.safeJsonParse(lib_types_2.Utils.safeStringify(cycle));
    delete cycleCopy.marker;
    delete cycleCopy.certificates;
    cycleCopy.nodeListHash = '';
    cycleCopy.archiverListHash = '';
    cycleCopy.standbyNodeListHash = '';
    cycleCopy.joinedConsensors.forEach((jc) => (jc.syncingTimestamp = 0));
    return cycleCopy;
}
//# sourceMappingURL=Data.js.map