import * as State from './State';
import { P2P as P2PTypes } from '@shardeum-foundation/lib-types';
import { SignedObject } from '@shardeum-foundation/lib-crypto-utils';
export declare enum NodeStatus {
    STANDBY = "standby",
    ACTIVE = "active",
    SYNCING = "syncing"
}
export interface ConsensusNodeInfo {
    ip: string;
    port: number;
    publicKey: string;
    id?: string;
    externalIp?: string;
    externalPort?: number;
    foundationNode?: boolean;
}
export interface ConsensusNodeListResponse {
    nodeList: ConsensusNodeInfo[];
}
export interface JoinedConsensor extends ConsensusNodeInfo {
    cycleJoined: string;
    counterRefreshed: number;
    id: string;
}
export declare const byAscendingPublicKey: (a: State.ArchiverNodeInfo, b: State.ArchiverNodeInfo) => number;
export declare const fromP2PTypesJoinedConsensor: (joinedConsensor: P2PTypes.JoinTypes.JoinedConsensor) => JoinedConsensor;
export declare const fromP2PTypesNode: (node: P2PTypes.NodeListTypes.Node) => JoinedConsensor;
export declare const byPublicKey: Map<string, ConsensusNodeInfo>;
export declare let activeListByIdSorted: ConsensusNodeInfo[];
export declare let foundFirstNode: boolean;
export type SignedNodeList = {
    nodeList: ConsensusNodeInfo[];
} & SignedObject;
export declare const activeNodescache: Map<string, SignedNodeList>;
export declare const fullNodesCache: Map<string, SignedNodeList>;
export declare const cacheUpdatedTimes: Map<string, number>;
export declare const realUpdatedTimes: Map<string, number>;
export declare function isEmpty(): boolean;
type Node = ConsensusNodeInfo | JoinedConsensor;
export declare function addNodes(status: NodeStatus, nodes: Node[]): void;
export declare function refreshNodes(status: NodeStatus, nodes: ConsensusNodeInfo[] | JoinedConsensor[]): void;
export declare function removeNodes(publicKeys: string[]): void;
export declare const addStandbyNodes: (nodes: ConsensusNodeInfo[]) => void;
export declare const removeStandbyNodes: (publicKeys: string[]) => void;
export declare function setStatus(status: Exclude<NodeStatus, NodeStatus.STANDBY>, publicKeys: string[]): void;
export declare function getFirstNode(): ConsensusNodeInfo | undefined;
export declare function getActiveList(id_sorted?: boolean): ConsensusNodeInfo[];
export declare function getActiveNodeCount(): number;
/**
 * Check the cache for the node list, if it's hot, return it. Otherwise,
 * rebuild the cache and return the node list.
 */
export declare const getCachedNodeList: () => SignedNodeList;
export declare const getCachedFullNodeList: (activeOnly: boolean, syncingOnly: boolean, standbyOnly: boolean) => SignedNodeList;
export declare function getActiveNodeListFromArchiver(archiver: State.ArchiverNodeInfo): Promise<ConsensusNodeInfo[]>;
export declare function getRandomActiveNodes(node_count?: number): ConsensusNodeInfo[];
export declare function getSyncingList(): ConsensusNodeInfo[];
export declare function getStandbyList(): ConsensusNodeInfo[];
export declare function getNodeInfoById(id: string): ConsensusNodeInfo | undefined;
export declare function changeNodeListInRestore(): void;
/** Resets/Cleans all the NodeList associated Maps and Array variables/caches */
export declare function clearNodeLists(): void;
export declare function toggleFirstNode(): void;
export {};
