"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
exports.toggleFirstNode = exports.clearNodeLists = exports.changeNodeListInRestore = exports.getNodeInfoById = exports.getStandbyList = exports.getSyncingList = exports.getRandomActiveNodes = exports.getActiveNodeListFromArchiver = exports.getCachedFullNodeList = exports.getCachedNodeList = exports.getActiveNodeCount = exports.getActiveList = exports.getFirstNode = exports.setStatus = exports.removeStandbyNodes = exports.addStandbyNodes = exports.removeNodes = exports.refreshNodes = exports.addNodes = exports.isEmpty = exports.realUpdatedTimes = exports.cacheUpdatedTimes = exports.fullNodesCache = exports.activeNodescache = exports.foundFirstNode = exports.activeListByIdSorted = exports.byPublicKey = exports.fromP2PTypesNode = exports.fromP2PTypesJoinedConsensor = exports.byAscendingPublicKey = exports.NodeStatus = void 0;
const State = require("./State");
const P2P = require("./P2P");
const Utils = require("./Utils");
const Logger = require("./Logger");
const Config_1 = require("./Config");
const Crypto = require("./Crypto");
const lib_types_1 = require("@shardeum-foundation/lib-types");
// TYPES
var NodeStatus;
(function (NodeStatus) {
    NodeStatus["STANDBY"] = "standby";
    NodeStatus["ACTIVE"] = "active";
    NodeStatus["SYNCING"] = "syncing";
})(NodeStatus = exports.NodeStatus || (exports.NodeStatus = {}));
const byAscendingNodeId = (a, b) => (a.id > b.id ? 1 : -1);
const byAscendingPublicKey = (a, b) => a.publicKey > b.publicKey ? 1 : -1;
exports.byAscendingPublicKey = byAscendingPublicKey;
const fromP2PTypesJoinedConsensor = (joinedConsensor) => ({
    ip: joinedConsensor.internalIp,
    port: joinedConsensor.internalPort,
    publicKey: joinedConsensor.publicKey,
    id: joinedConsensor.id,
    externalIp: joinedConsensor.externalIp,
    externalPort: joinedConsensor.externalPort,
    cycleJoined: joinedConsensor.cycleJoined,
    counterRefreshed: joinedConsensor.counterRefreshed,
});
exports.fromP2PTypesJoinedConsensor = fromP2PTypesJoinedConsensor;
const fromP2PTypesNode = (node) => ({
    ip: node.internalIp,
    port: node.internalPort,
    publicKey: node.publicKey,
    id: node.id,
    externalIp: node.externalIp,
    externalPort: node.externalPort,
    cycleJoined: node.cycleJoined,
    counterRefreshed: node.counterRefreshed,
});
exports.fromP2PTypesNode = fromP2PTypesNode;
// STATE
const standbyList = new Map();
const syncingList = new Map();
const activeList = new Map();
// Map to get node public key by node Id
const byId = new Map();
// Map to get node info by public key, stores syncing and active nodes
exports.byPublicKey = new Map();
// Array of active nodes sorted by id
exports.activeListByIdSorted = [];
exports.foundFirstNode = false;
exports.activeNodescache = new Map();
exports.fullNodesCache = new Map();
exports.cacheUpdatedTimes = new Map();
exports.realUpdatedTimes = new Map();
// METHODS
function isEmpty() {
    return exports.byPublicKey.size === 0;
}
exports.isEmpty = isEmpty;
function addNodes(status, nodes) {
    if (nodes.length === 0)
        return;
    Logger.mainLogger.debug(`Adding ${status} nodes to the list`, nodes.length, nodes);
    for (const node of nodes) {
        // If node not in lists, add it
        // eslint-disable-next-line security/detect-object-injection
        if (!exports.byPublicKey.has(node.publicKey)) {
            const key = node.publicKey;
            switch (status) {
                case NodeStatus.SYNCING:
                    if (standbyList.has(key))
                        standbyList.delete(key);
                    if (activeList.has(key)) {
                        activeList.delete(key);
                        exports.activeListByIdSorted = exports.activeListByIdSorted.filter((node) => node.publicKey !== key);
                    }
                    if (syncingList.has(key))
                        break;
                    syncingList.set(node.publicKey, node);
                    break;
                case NodeStatus.ACTIVE:
                    if (standbyList.has(key))
                        standbyList.delete(key);
                    if (syncingList.has(key))
                        syncingList.delete(key);
                    if (activeList.has(key))
                        break;
                    activeList.set(node.publicKey, node);
                    Utils.insertSorted(exports.activeListByIdSorted, node, byAscendingNodeId);
                    // Logger.mainLogger.debug(
                    //   'activeListByIdSorted',
                    //   activeListByIdSorted.map((node) => node.id)
                    // )
                    break;
            }
            /* eslint-disable security/detect-object-injection */
            exports.byPublicKey.set(node.publicKey, node);
            /* eslint-enable security/detect-object-injection */
        }
        // If an id is given, update its id
        if (node.id) {
            const entry = exports.byPublicKey.get(node.publicKey);
            if (entry) {
                entry.id = node.id;
                byId.set(node.id, node.publicKey);
            }
        }
    }
}
exports.addNodes = addNodes;
function refreshNodes(status, nodes) {
    if (nodes.length === 0)
        return;
    Logger.mainLogger.debug('Refreshing nodes', nodes.length, nodes);
    for (const node of nodes) {
        // If node not in lists, add it
        // eslint-disable-next-line security/detect-object-injection
        if (!exports.byPublicKey.has(node.publicKey)) {
            Logger.mainLogger.debug('adding new node during refresh', node.publicKey);
            switch (status) {
                case NodeStatus.SYNCING:
                    syncingList.set(node.publicKey, node);
                    break;
                case NodeStatus.ACTIVE:
                    activeList.set(node.publicKey, node);
                    Utils.insertSorted(exports.activeListByIdSorted, node, byAscendingNodeId);
                    // Logger.mainLogger.debug(
                    //   'activeListByIdSorted',
                    //   activeListByIdSorted.map((node) => node.id)
                    // )
                    break;
            }
            exports.byPublicKey.set(node.publicKey, node);
            // eslint-disable-next-line security/detect-object-injection
        }
        // If an id is given, update its id
        if (node.id) {
            const entry = exports.byPublicKey.get(node.publicKey);
            if (entry) {
                entry.id = node.id;
                byId.set(node.id, node.publicKey);
            }
        }
    }
}
exports.refreshNodes = refreshNodes;
function removeNodes(publicKeys) {
    if (publicKeys.length > 0)
        Logger.mainLogger.debug('Removing nodes', publicKeys);
    // Efficiently remove nodes from nodelist
    for (const key of publicKeys) {
        // eslint-disable-next-line security/detect-object-injection
        if (!exports.byPublicKey.has(key)) {
            console.warn(`removeNodes: publicKey ${key} not in nodelist`);
            continue;
        }
        /* eslint-disable security/detect-object-injection */
        syncingList.delete(key);
        activeList.delete(key);
        standbyList.delete(key);
        const id = exports.byPublicKey.get(key).id;
        exports.activeListByIdSorted = exports.activeListByIdSorted.filter((node) => node.id !== exports.byPublicKey.get(key).id);
        byId.delete(id);
        exports.byPublicKey.delete(key);
        /* eslint-enable security/detect-object-injection */
    }
}
exports.removeNodes = removeNodes;
const addStandbyNodes = (nodes) => {
    if (nodes.length === 0)
        return;
    Logger.mainLogger.debug('Adding standby nodes to the list', nodes.length, nodes);
    for (const node of nodes) {
        if (standbyList.has(node.publicKey))
            continue;
        standbyList.set(node.publicKey, node);
    }
};
exports.addStandbyNodes = addStandbyNodes;
const removeStandbyNodes = (publicKeys) => {
    if (publicKeys.length > 0)
        Logger.mainLogger.debug('Removing standby nodes', publicKeys);
    for (const key of publicKeys) {
        if (standbyList.has(key))
            standbyList.delete(key);
    }
};
exports.removeStandbyNodes = removeStandbyNodes;
function setStatus(status, publicKeys) {
    if (publicKeys.length === 0)
        return;
    Logger.mainLogger.debug(`Updating status ${status} for nodes`, publicKeys);
    for (const key of publicKeys) {
        // eslint-disable-next-line security/detect-object-injection
        const node = exports.byPublicKey.get(key);
        if (node === undefined) {
            console.warn(`setStatus: publicKey ${key} not in nodelist`);
            continue;
        }
        switch (status) {
            case NodeStatus.SYNCING:
                if (standbyList.has(key))
                    standbyList.delete(key);
                if (activeList.has(key)) {
                    activeList.delete(key);
                    exports.activeListByIdSorted = exports.activeListByIdSorted.filter((node) => node.publicKey !== key);
                }
                if (syncingList.has(key))
                    break;
                syncingList.set(key, node);
                break;
            case NodeStatus.ACTIVE:
                if (standbyList.has(key))
                    standbyList.delete(key);
                if (syncingList.has(key))
                    syncingList.delete(key);
                if (activeList.has(key))
                    break;
                activeList.set(key, node);
                Utils.insertSorted(exports.activeListByIdSorted, node, byAscendingNodeId);
                // Logger.mainLogger.debug(
                //   'activeListByIdSorted',
                //   activeListByIdSorted.map((node) => node.id)
                // )
                break;
        }
    }
}
exports.setStatus = setStatus;
function getFirstNode() {
    return exports.byPublicKey?.values()?.next()?.value;
}
exports.getFirstNode = getFirstNode;
function getActiveList(id_sorted = true) {
    if (id_sorted)
        return exports.activeListByIdSorted;
    return [...activeList.values()];
}
exports.getActiveList = getActiveList;
function getActiveNodeCount() {
    return activeList.size;
}
exports.getActiveNodeCount = getActiveNodeCount;
/**
 * Check the cache for the node list, if it's hot, return it. Otherwise,
 * rebuild the cache and return the node list.
 */
const getCachedNodeList = () => {
    const cacheUpdatedTime = exports.cacheUpdatedTimes.get('/nodelist');
    const realUpdatedTime = exports.realUpdatedTimes.get('/nodelist');
    const bucketCacheKey = (index) => `/nodelist/${index}`;
    if (cacheUpdatedTime && realUpdatedTime && cacheUpdatedTime > realUpdatedTime) {
        // cache is hot, send cache
        const randomIndex = Math.floor(Math.random() * Config_1.config.N_RANDOM_NODELIST_BUCKETS);
        const cachedNodeList = exports.activeNodescache.get(bucketCacheKey(randomIndex));
        return cachedNodeList;
    }
    // cache is cold, remake cache
    const nodeCount = Math.min(Config_1.config.N_NODELIST, getActiveNodeCount());
    for (let index = 0; index < Config_1.config.N_RANDOM_NODELIST_BUCKETS; index++) {
        // If we dont have any active nodes, send back the first node in our list
        const nodeList = nodeCount < 1 ? (isEmpty() ? [] : [getFirstNode()]) : getRandomActiveNodes(nodeCount);
        const sortedNodeList = [...nodeList].sort(byAscendingNodeId);
        const signedSortedNodeList = Crypto.sign({
            nodeList: sortedNodeList,
        });
        // Update cache
        exports.activeNodescache.set(bucketCacheKey(index), signedSortedNodeList);
    }
    // Update cache timestamps
    if (exports.realUpdatedTimes.get('/nodelist') === undefined) {
        // This gets set when the list of nodes changes. For the first time, set to a large value
        exports.realUpdatedTimes.set('/nodelist', Date.now());
    }
    exports.cacheUpdatedTimes.set('/nodelist', Date.now());
    const nodeList = exports.activeNodescache.get(bucketCacheKey(0));
    return nodeList;
};
exports.getCachedNodeList = getCachedNodeList;
const getCachedFullNodeList = (activeOnly, syncingOnly, standbyOnly) => {
    const cacheUpdatedTime = exports.cacheUpdatedTimes.get('/full-nodelist');
    const realUpdatedTime = exports.realUpdatedTimes.get('/full-nodelist');
    if (cacheUpdatedTime && realUpdatedTime && cacheUpdatedTime > realUpdatedTime) {
        // cache is hot, send cache
        if (activeOnly)
            return exports.fullNodesCache.get('/full-active-nodelist');
        if (syncingOnly)
            return exports.fullNodesCache.get('/full-syncing-nodelist');
        if (standbyOnly)
            return exports.fullNodesCache.get('/full-standby-nodelist');
        return exports.fullNodesCache.get('/full-active-syncing-nodelist');
    }
    // cache is cold, remake cache
    const activeNodeList = getActiveList(false);
    const syncingNodeList = getSyncingList();
    const standbyNodeList = getStandbyList();
    const activeSyncingNodeList = activeNodeList.concat(syncingNodeList); // active + syncing
    exports.fullNodesCache.set('/full-active-nodelist', Crypto.sign({ nodeList: activeNodeList }));
    exports.fullNodesCache.set('/full-syncing-nodelist', Crypto.sign({ nodeList: syncingNodeList }));
    exports.fullNodesCache.set('/full-standby-nodelist', Crypto.sign({ nodeList: standbyNodeList }));
    exports.fullNodesCache.set('/full-active-syncing-nodelist', Crypto.sign({ nodeList: activeSyncingNodeList }));
    // Update cache timestamps
    if (exports.realUpdatedTimes.get('/full-nodelist') === undefined) {
        // This gets set when the list of nodes changes. For the first time, set to a large value
        exports.realUpdatedTimes.set('/full-nodelist', Date.now());
    }
    exports.cacheUpdatedTimes.set('/full-nodelist', Date.now());
    if (activeOnly)
        return exports.fullNodesCache.get('/full-active-nodelist');
    if (syncingOnly)
        return exports.fullNodesCache.get('/full-syncing-nodelist');
    if (standbyOnly)
        return exports.fullNodesCache.get('/full-standby-nodelist');
    return exports.fullNodesCache.get('/full-active-syncing-nodelist');
};
exports.getCachedFullNodeList = getCachedFullNodeList;
async function getActiveNodeListFromArchiver(archiver) {
    const response = (await P2P.getJson(`http://${archiver.ip}:${archiver.port}/nodelist`));
    Logger.mainLogger.debug('response', `http://${archiver.ip}:${archiver.port}/nodelist`, response);
    if (response && response.nodeList && response.nodeList.length > 0) {
        const isResponseVerified = Crypto.verify(response);
        const isFromActiveArchiver = State.activeArchivers.some((archiver) => archiver.publicKey === response.sign.owner);
        if (!isResponseVerified || !isFromActiveArchiver) {
            Logger.mainLogger.debug(`Fail to verify the response from the archiver ${archiver.ip}:${archiver.port}`);
            return [];
        }
        return response.nodeList;
    }
    else
        Logger.mainLogger.debug(`Fail To get nodeList from the archiver ${archiver.ip}:${archiver.port}`);
    return [];
}
exports.getActiveNodeListFromArchiver = getActiveNodeListFromArchiver;
function getRandomActiveNodes(node_count = 1) {
    const nodeList = getActiveList();
    if (node_count <= 1 || node_count > nodeList.length)
        return Utils.getRandomItemFromArr(nodeList, Config_1.config.N_NODE_REJECT_PERCENT);
    return Utils.getRandomItemFromArr(nodeList, Config_1.config.N_NODE_REJECT_PERCENT, node_count);
}
exports.getRandomActiveNodes = getRandomActiveNodes;
function getSyncingList() {
    return [...syncingList.values()];
}
exports.getSyncingList = getSyncingList;
function getStandbyList() {
    return [...standbyList.values()];
}
exports.getStandbyList = getStandbyList;
function getNodeInfoById(id) {
    const pk = byId.get(id);
    return pk ? exports.byPublicKey.get(pk) : undefined;
}
exports.getNodeInfoById = getNodeInfoById;
function changeNodeListInRestore() {
    if (activeList.size === 0)
        return;
    // change the active status nodes to syncing status in all the nodelist
    const activatedPublicKeys = exports.activeListByIdSorted.map((node) => node.publicKey);
    Logger.mainLogger.debug('changeNodeListInRestore: Changing node list in restore mode: ', lib_types_1.Utils.safeStringify(activatedPublicKeys));
    setStatus(NodeStatus.SYNCING, activatedPublicKeys);
}
exports.changeNodeListInRestore = changeNodeListInRestore;
/** Resets/Cleans all the NodeList associated Maps and Array variables/caches */
function clearNodeLists() {
    try {
        exports.activeNodescache.clear();
        exports.fullNodesCache.clear();
        activeList.clear();
        syncingList.clear();
        exports.realUpdatedTimes.clear();
        exports.cacheUpdatedTimes.clear();
        byId.clear();
        exports.byPublicKey.clear();
        exports.activeListByIdSorted = [];
    }
    catch (e) {
        Logger.mainLogger.error('Error thrown in clearNodeListCache', e);
    }
}
exports.clearNodeLists = clearNodeLists;
function toggleFirstNode() {
    exports.foundFirstNode = !exports.foundFirstNode;
    Logger.mainLogger.debug('foundFirstNode', exports.foundFirstNode);
}
exports.toggleFirstNode = toggleFirstNode;
//# sourceMappingURL=NodeList.js.map