"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
exports.get = exports.getJson = exports.postJson = exports.createArchiverLeaveRequest = exports.createArchiverActiveRequest = exports.createArchiverJoinRequest = exports.RequestTypes = void 0;
const State = require("./State");
const Crypto = require("./Crypto");
const lib_types_1 = require("@shardeum-foundation/lib-types");
const customHttpFunctions_1 = require("./utils/customHttpFunctions");
// eslint-disable-next-line @typescript-eslint/no-var-requires
const { version } = require('../package.json');
var RequestTypes;
(function (RequestTypes) {
    RequestTypes["JOIN"] = "JOIN";
    RequestTypes["ACTIVE"] = "ACTIVE";
    RequestTypes["LEAVE"] = "LEAVE";
})(RequestTypes = exports.RequestTypes || (exports.RequestTypes = {}));
function createArchiverJoinRequest() {
    const joinRequest = {
        nodeInfo: State.getNodeInfo(),
        appData: { version },
        requestType: RequestTypes.JOIN,
        requestTimestamp: Date.now(),
    };
    return Crypto.sign(joinRequest);
}
exports.createArchiverJoinRequest = createArchiverJoinRequest;
function createArchiverActiveRequest() {
    const activeRequest = {
        nodeInfo: State.getNodeInfo(),
        requestType: RequestTypes.ACTIVE,
    };
    return Crypto.sign(activeRequest);
}
exports.createArchiverActiveRequest = createArchiverActiveRequest;
function createArchiverLeaveRequest() {
    const leaveRequest = {
        nodeInfo: State.getNodeInfo(),
        requestType: RequestTypes.LEAVE,
        requestTimestamp: Date.now(),
    };
    return Crypto.sign(leaveRequest);
}
exports.createArchiverLeaveRequest = createArchiverLeaveRequest;
async function postJson(url, body, timeoutInSecond = 5) {
    try {
        const res = await (0, customHttpFunctions_1.customFetch)(url, {
            method: 'post',
            body: lib_types_1.Utils.safeStringify(body),
            headers: { 'Content-Type': 'application/json' },
            timeout: timeoutInSecond * 1000,
        });
        if (res.ok) {
            const text = await res.text();
            try {
                return lib_types_1.Utils.safeJsonParse(text);
            }
            catch (parseError) {
                console.warn(`getJson failed: invalid JSON response url: ${url} parseError: ${parseError}`);
                return null;
            }
        }
        else {
            console.warn('postJson failed: got bad response');
            console.warn(res.headers);
            console.warn(res.statusText);
            console.warn(await res.text());
            return null;
        }
    }
    catch (err) {
        console.warn('postJson failed: could not reach host');
        console.warn(err);
        return null;
    }
}
exports.postJson = postJson;
async function getJson(url, timeoutInSecond = 5) {
    try {
        const res = await get(url, timeoutInSecond, {
            headers: { 'Content-Type': 'application/json' },
        });
        if (res.ok) {
            const text = await res.text();
            try {
                return lib_types_1.Utils.safeJsonParse(text);
            }
            catch (parseError) {
                console.warn(`getJson failed: invalid JSON response url: ${url} parseError: ${parseError}`);
                return null;
            }
        }
        else {
            console.warn('getJson failed: got bad response');
            console.warn(url);
            console.warn(res.headers);
            console.warn(res.statusText);
            console.warn(await res.text());
            return null;
        }
    }
    catch (err) {
        console.warn('getJson failed: could not reach host');
        console.warn(err);
        return null;
    }
}
exports.getJson = getJson;
async function get(url, timeoutInSecond = 20, opts) {
    return (0, customHttpFunctions_1.customFetch)(url, {
        method: 'get',
        timeout: timeoutInSecond * 1000,
        ...opts,
    });
}
exports.get = get;
//# sourceMappingURL=P2P.js.map