"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
exports.receiptCheckpointManager = exports.ReceiptRadixDigestTally = exports.ReceiptCheckpointBucket = exports.ReceiptCheckpointRadixDigest = exports.ReceiptCheckpointRadixEntry = exports.calculateBucketID = exports.ReceiptCheckpointData = void 0;
const CheckpointData_1 = require("./CheckpointData");
const receipts_1 = require("../dbstore/receipts");
const Logger = require("../Logger");
const Crypto = require("../Crypto");
const verifyAppReceiptData_1 = require("../shardeum/verifyAppReceiptData");
const lib_types_1 = require("@shardeum-foundation/lib-types");
class ReceiptCheckpointData extends CheckpointData_1.CheckpointData {
    constructor(receipt) {
        const receiptHash = Crypto.hash(lib_types_1.Utils.safeStringify(receipt)).toLowerCase();
        super(receiptHash.substring(0, 2), // address (first 2 chars)
        receipt.tx.timestamp, // timestamp
        receiptHash, // hash
        CheckpointData_1.CheckpointType.Receipt, // class type 2 for receipts
        receipt // data
        );
    }
}
exports.ReceiptCheckpointData = ReceiptCheckpointData;
function calculateBucketID(receipt) {
    if (!receipt || receipt.tx.timestamp === undefined) {
        Logger.mainLogger.error('Invalid receipt data');
        throw new Error('Invalid receipt data');
    }
    return receipt.cycle.toString();
}
exports.calculateBucketID = calculateBucketID;
//Represents a single radix entry in a bucket
class ReceiptCheckpointRadixEntry extends CheckpointData_1.CheckpointRadixEntry {
    constructor(radix) {
        super(radix);
    }
}
exports.ReceiptCheckpointRadixEntry = ReceiptCheckpointRadixEntry;
//Represents a single radix entry in a bucket
class ReceiptCheckpointRadixDigest extends CheckpointData_1.CheckpointRadixDigest {
    constructor(radix, hash, itemCount) {
        super(radix, hash, itemCount);
    }
}
exports.ReceiptCheckpointRadixDigest = ReceiptCheckpointRadixDigest;
//Represents a single bucket in the system
class ReceiptCheckpointBucket extends CheckpointData_1.CheckpointBucket {
    constructor(startTime, endTime, bucketID, validateData, updateData) {
        super(startTime, endTime, bucketID, validateData, updateData, CheckpointData_1.CheckpointType.Receipt);
    }
    async update(currentTime) {
        // Call parent update first
        super.update(currentTime);
    }
}
exports.ReceiptCheckpointBucket = ReceiptCheckpointBucket;
class ReceiptCheckpointManager extends CheckpointData_1.CheckpointBucketManager {
    constructor() {
        const persistenceCallbacks = {
            validateData: validateData,
            updateData: updateData,
        };
        super(persistenceCallbacks, CheckpointData_1.CheckpointType.Receipt);
    }
    static getInstance() {
        if (!ReceiptCheckpointManager.instance) {
            ReceiptCheckpointManager.instance = new ReceiptCheckpointManager();
        }
        return ReceiptCheckpointManager.instance;
    }
}
//Represents a tally of all radix entries in the system
class ReceiptRadixDigestTally extends CheckpointData_1.RadixDigestTally {
    constructor(radix) {
        super(radix);
    }
}
exports.ReceiptRadixDigestTally = ReceiptRadixDigestTally;
async function validateData(data) {
    try {
        const verifyHash = Crypto.hash(lib_types_1.Utils.safeStringify(data.d)).toLowerCase();
        if (verifyHash !== data.h) {
            return false;
        }
        const appValidation = await (0, verifyAppReceiptData_1.verifyAppReceiptData)(data.d, null, [], []);
        return appValidation.valid;
    }
    catch (err) {
        Logger.mainLogger.error('ValidateData failed:', err);
        return false;
    }
}
async function updateData(data) {
    try {
        // Insert/Update into checkpoint_data table
        const receipt = data.d;
        // Avg entry size is about 41000 bytes
        await (0, receipts_1.insertReceipt)(receipt, false);
    }
    catch (err) {
        Logger.mainLogger.error('Failed to store receipt checkpoint data:', err);
        throw err;
    }
}
// Export the singleton instance
exports.receiptCheckpointManager = ReceiptCheckpointManager.getInstance();
//# sourceMappingURL=ReceiptData.js.map