"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
exports.stringifyReduce = exports.makeShortHash = void 0;
const lib_types_1 = require("@shardeum-foundation/lib-types");
const makeShortHash = (x, n = 4) => {
    if (!x) {
        return x;
    }
    if (x.length > 63) {
        if (x.length === 64) {
            return x.slice(0, n) + 'x' + x.slice(63 - n);
        }
        else if (x.length === 128) {
            return x.slice(0, n) + 'xx' + x.slice(127 - n);
        }
        else if (x.length === 192) {
            return x.slice(0, n) + 'xx' + x.slice(191 - n);
        }
    }
    return x;
};
exports.makeShortHash = makeShortHash;
const objToString = Object.prototype.toString;
const objKeys = ((obj) => {
    const keys = [];
    // tslint:disable-next-line: forin
    for (const name in obj) {
        keys.push(name);
    }
    return keys;
}) || Object.keys;
const stringifyReduce = (val, isArrayProp) => {
    let i;
    let max;
    let str;
    let keys;
    let key;
    let propVal;
    let toStr;
    if (val === true) {
        return 'true';
    }
    if (val === false) {
        return 'false';
    }
    switch (typeof val) {
        case 'object':
            if (val === null) {
                return null;
            }
            // not used, don't compile for object
            // else if (val.toJSON && typeof val.toJSON === 'function') {
            //   return stringifyReduce(val.toJSON(), isArrayProp)
            // }
            else if (val instanceof Map) {
                const mapContainer = {
                    dataType: 'stringifyReduce_map_2_array',
                    value: Array.from(val.entries()), // or with spread: value: [...originalObject]
                };
                return (0, exports.stringifyReduce)(mapContainer);
            }
            else {
                toStr = objToString.call(val);
                if (toStr === '[object Array]') {
                    str = '[';
                    max = val.length - 1;
                    for (i = 0; i < max; i++) {
                        // eslint-disable-next-line security/detect-object-injection
                        str += (0, exports.stringifyReduce)(val[i], true) + ',';
                    }
                    if (max > -1) {
                        // eslint-disable-next-line security/detect-object-injection
                        str += (0, exports.stringifyReduce)(val[i], true);
                    }
                    return str + ']';
                }
                else if (toStr === '[object Object]') {
                    // only object is left
                    keys = objKeys(val).sort();
                    max = keys.length;
                    str = '';
                    i = 0;
                    while (i < max) {
                        // eslint-disable-next-line security/detect-object-injection
                        key = keys[i];
                        // eslint-disable-next-line security/detect-object-injection
                        propVal = (0, exports.stringifyReduce)(val[key], false);
                        if (propVal !== undefined) {
                            if (str) {
                                str += ',';
                            }
                            str += lib_types_1.Utils.safeStringify(key) + ':' + propVal;
                        }
                        i++;
                    }
                    return '{' + str + '}';
                }
                else {
                    return lib_types_1.Utils.safeStringify(val);
                }
            }
        case 'function':
        case 'undefined':
            return isArrayProp ? null : undefined;
        case 'string': {
            const reduced = (0, exports.makeShortHash)(val);
            return lib_types_1.Utils.safeStringify(reduced);
        }
        default: {
            const n = Number(val);
            return isFinite(n) ? n.toString() : null;
        }
    }
};
exports.stringifyReduce = stringifyReduce;
//TODO not used
// export const replacer = (key: any, value: any) => {
//   const originalObject = value // this[key]
//   if (originalObject instanceof Map) {
//     return {
//       dataType: 'stringifyReduce_map_2_array',
//       value: Array.from(originalObject.entries()), // or with spread: value: [...originalObject]
//     }
//   } else {
//     return value
//   }
// }
//TODO not used
// export const reviver = (key: any, value: any) => {
//   if (typeof value === 'object' && value !== null) {
//     if (value.dataType === 'stringifyReduce_map_2_array') {
//       return new Map(value.value)
//     }
//   }
//   return value
// }
//TODO not used
// export const reviverExpander = (key: string, value: any) => {
//   if (typeof value === 'object' && value !== null) {
//     if (value.dataType === 'stringifyReduce_map_2_array') {
//       return new Map(value.value)
//     }
//   }
//   if (typeof value === 'string' && value.length === 10 && value[4] === 'x') {
//     const res = value.slice(0, 4) + '0'.repeat(55) + value.slice(5, 5 + 5)
//     return res
//   }
//   return value
// }
//# sourceMappingURL=StringifyReduce.js.map