export interface ArchiverRegistrationLog {
    archiverId: string;
    timestamp: number;
    validators: {
        discovered: number;
        connected: number;
    };
    state: 'REGISTERING' | 'REGISTERED' | 'ERROR';
}
export interface ValidatorConnectionLog {
    validatorId: string;
    archiverId: string;
    timestamp: number;
    status: 'CONNECTING' | 'CONNECTED' | 'ERROR';
    handshake: {
        success: boolean;
        duration: number;
        error?: string;
    };
}
export interface DataSyncLog {
    sourceArchiver: string;
    targetArchiver: string;
    cycle: number;
    dataType: 'VALIDATOR_LIST' | 'CYCLE_RECORD' | 'TX_LIST' | 'STANDBY_LIST' | 'ARCHIVER_LIST';
    dataHash: string;
    status: 'STARTED' | 'IN_PROGRESS' | 'COMPLETE' | 'ERROR';
    metrics: {
        duration: number;
        dataSize: number;
    };
    error?: string;
    operationId?: string;
}
export type OperationType = 'CYCLE_SYNC' | 'VALIDATOR_SYNC' | 'TX_SYNC' | 'REGISTRATION' | 'CONNECTION';
export declare class ArchiverLogging {
    static generateOperationId(): string;
    static logArchiverRegistration(log: ArchiverRegistrationLog): void;
    static logValidatorConnection(log: ValidatorConnectionLog): void;
    static logDataSync(log: DataSyncLog): void;
}
