"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
const path_1 = require("path");
const cron = require("node-cron");
const dbstore = require("./dbstore");
const txDigesterDB = require("./txDigester/index");
const txDigestFunctions = require("./txDigester/txDigestFunctions");
const Config_1 = require("./Config");
const CycleDB = require("./dbstore/cycles");
const Crypto = require("./Crypto");
const lib_types_1 = require("@shardeum-foundation/lib-types");
const fs_1 = require("fs");
const path_2 = require("path");
const Logger = require("./Logger");
const saveConsoleOutput_1 = require("./saveConsoleOutput");
const customHttpFunctions_1 = require("./utils/customHttpFunctions");
const configFile = (0, path_1.join)(process.cwd(), 'archiver-config.json');
const start = async () => {
    (0, Config_1.overrideDefaultConfig)(configFile);
    const hashKey = Config_1.config.ARCHIVER_HASH_KEY;
    Crypto.setCryptoHashKey(hashKey);
    let logsConfig;
    try {
        logsConfig = lib_types_1.Utils.safeJsonParse((0, fs_1.readFileSync)((0, path_2.resolve)(__dirname, '../archiver-log.json'), 'utf8'));
    }
    catch (err) {
        console.log('Failed to parse archiver log file:', err);
    }
    const logDir = `${Config_1.config.ARCHIVER_LOGS}/txDigester`;
    const baseDir = '.';
    logsConfig.dir = logDir;
    Logger.initLogger(baseDir, logsConfig);
    if (logsConfig.saveConsoleOutput) {
        (0, saveConsoleOutput_1.startSaving)((0, path_1.join)(baseDir, logsConfig.dir));
    }
    await dbstore.initializeDB(Config_1.config);
    await txDigesterDB.initializeDB(Config_1.config);
    const ARCHIVER_STATUS_CHECK_URL = `http://${Config_1.config.ARCHIVER_IP}:${Config_1.config.ARCHIVER_PORT}/status`;
    cron.schedule(Config_1.config.txDigest.txCronSchedule, async () => {
        console.log('Running cron task....');
        console.log('Checking archiver status....');
        const archiverStatusResp = await (0, customHttpFunctions_1.customAxios)().get(ARCHIVER_STATUS_CHECK_URL);
        const isArchiverActive = archiverStatusResp.data.status.isActive;
        console.log('isArchiverActive: ', isArchiverActive);
        if (isArchiverActive) {
            const lastProcessedTxDigest = await txDigestFunctions.getLastProcessedTxDigest();
            console.log('lastProcessedTxDigest by txDigester: ', lastProcessedTxDigest);
            const lastCheckedCycle = lastProcessedTxDigest ? lastProcessedTxDigest.cycleEnd : -1;
            console.log('lastCheckedCycle by txDigester: ', lastCheckedCycle);
            const latestCycleRecords = await CycleDB.queryLatestCycleRecords(1);
            const latestCycleCounter = latestCycleRecords.length > 0 ? latestCycleRecords[0].counter : -1;
            console.log('latestCycleCounter reported by Archiver: ', latestCycleCounter);
            const latestSyncedCycleCounter = latestCycleCounter - Config_1.config.txDigest.syncDelay;
            if (latestSyncedCycleCounter - lastCheckedCycle >= Config_1.config.txDigest.cycleDiff) {
                await txDigestFunctions.processAndInsertTxDigests(lastCheckedCycle + 1, latestSyncedCycleCounter);
            }
        }
        else {
            console.log('Archiver is not active. Skipping txDigest processing....');
        }
    });
};
start();
//# sourceMappingURL=txDigester.js.map