"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
exports.getVerifyFunction = exports.initializeSerialization = exports.addSchema = void 0;
const Ajv = require("ajv");
const ajv = new Ajv();
const schemaMap = new Map();
const verifyFunctions = new Map();
function addSchema(name, schema) {
    if (schemaMap.has(name)) {
        throw new Error(`error already registered ${name}`);
    }
    schemaMap.set(name, schema);
}
exports.addSchema = addSchema;
function initializeSerialization() {
    // Register each schema exactly once in AJV
    for (const [name, schema] of schemaMap.entries()) {
        ajv.addSchema(schema, name);
    }
}
exports.initializeSerialization = initializeSerialization;
function getVerifyFunction(name) {
    const existingFn = verifyFunctions.get(name);
    if (existingFn) {
        return existingFn;
    }
    const schema = schemaMap.get(name);
    if (!schema) {
        throw new Error(`error missing schema ${name}`);
    }
    const verifyFn = ajv.compile(schema);
    verifyFunctions.set(name, verifyFn);
    return verifyFn;
}
exports.getVerifyFunction = getVerifyFunction;
//# sourceMappingURL=SchemaHelpers.js.map