"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
exports.initWorkerProcess = void 0;
const Collector_1 = require("../Data/Collector");
const Config_1 = require("../Config");
const lib_types_1 = require("@shardeum-foundation/lib-types");
const initWorkerProcess = async () => {
    console.log(`Worker ${process.pid} started`);
    let lastActivity = Date.now();
    // Worker processes
    process.on('message', async ({ type, data }) => {
        switch (type) {
            case 'receipt-verification': {
                if (!data.stringifiedReceipt) {
                    console.error(`Worker ${process.pid} received invalid receipt for verification`, data);
                    return;
                }
                if (isNaN(data.requiredSignatures)) {
                    console.error(`Worker ${process.pid} received invalid requiredSignatures for verification`, data);
                    return;
                }
                const receipt = lib_types_1.Utils.safeJsonParse(data.stringifiedReceipt);
                // console.log(`Worker ${process.pid} verifying receipt`);
                let verificationResult = {
                    success: false,
                    failedReasons: [],
                    nestedCounterMessages: [],
                };
                try {
                    // console.log(`Worker process ${process.pid} is verifying receipt`, receipt.tx.txId, receipt.tx.timestamp)
                    // this code is not being used anymore, worker threads need to be enabled. Ignore for now.
                    verificationResult = await (0, Collector_1.verifyArchiverReceipt)(receipt);
                }
                catch (error) {
                    console.error(`Error in Worker ${process.pid} while verifying receipt`, error);
                    verificationResult.failedReasons.push('Error in Worker while verifying receipt');
                    verificationResult.nestedCounterMessages.push('Error in Worker while verifying receipt');
                }
                process.send({
                    type: 'receipt-verification',
                    data: {
                        txId: receipt.tx.txId,
                        timestamp: receipt.tx.timestamp,
                        verificationResult,
                    },
                });
                break;
            }
            default:
                console.log(`Worker ${process.pid} received unknown message type: ${type}`);
                console.log(data);
                break;
        }
        lastActivity = Date.now();
    });
    process.send({ type: 'child_ready' });
    setInterval(() => {
        // console.log(
        //   `lastActivityCheckTimeout: ${config.lastActivityCheckTimeout}, lastActivityCheckInterval: ${config.lastActivityCheckInterval}`
        // )
        if (Date.now() - lastActivity > Config_1.config.lastActivityCheckTimeout) {
            console.log(`Worker ${process.pid} is idle for more than 1 minute`);
            process.send({ type: 'child_close' });
        }
    }, Config_1.config.lastActivityCheckInterval);
};
exports.initWorkerProcess = initWorkerProcess;
process.on('uncaughtException', (error) => {
    console.error(`Uncaught Exception in Child Process: ${process.pid}`, error);
});
process.on('unhandledRejection', (reason, promise) => {
    console.error('Unhandled Rejection in Child Process:', promise, 'reason:', reason);
});
//# sourceMappingURL=index.js.map